@@ -11,7 +11,7 @@
 """
 
 # flake8: noqa: F401
-
+from .changeable import admin_change_password
 from .core import Security, RoleMixin, UserMixin, AnonymousUser, current_user
 from .datastore import (
     UserDatastore,
@@ -77,6 +77,7 @@ from .unified_signin import (
     UnifiedVerifyForm,
     us_send_security_token,
 )
+from .username_util import UsernameUtil
 from .utils import (
     FsJsonEncoder,
     SmsSenderBaseClass,
@@ -98,9 +99,10 @@ from .utils import (
     transform_url,
     uia_phone_mapper,
     uia_email_mapper,
+    uia_username_mapper,
     url_for_security,
     verify_password,
     verify_and_update_password,
 )
 
-__version__ = "4.0.1"
+__version__ = "4.1.1"
