@@ -1,4 +1,3 @@
-# -*- coding: utf-8 -*-
 """
     flask_security.datastore
     ~~~~~~~~~~~~~~~~~~~~~~~~
@@ -6,16 +5,23 @@
     This module contains an user datastore classes.
 
     :copyright: (c) 2012 by Matt Wright.
-    :copyright: (c) 2019-2020 by J. Christopher Wagner (jwag).
+    :copyright: (c) 2019-2021 by J. Christopher Wagner (jwag).
     :license: MIT, see LICENSE for more details.
 """
+import datetime
 import json
+import typing as t
 import uuid
 
-from .utils import config_value, get_identity_attributes, string_types
+from .utils import config_value
+
+if t.TYPE_CHECKING:  # pragma: no cover
+    import flask_sqlalchemy
+    import flask_mongoengine
+    import sqlalchemy.orm.scoping
 
 
-class Datastore(object):
+class Datastore:
     def __init__(self, db):
         self.db = db
 
@@ -115,7 +121,7 @@ class PonyDatastore(Datastore):
         model.delete()
 
 
-class UserDatastore(object):
+class UserDatastore:
     """Abstracted user datastore.
 
     :param user_model: A user model class definition
@@ -128,16 +134,16 @@ class UserDatastore(object):
         commit the transaction by calling datastore.commit().
     """
 
-    def __init__(self, user_model, role_model):
+    def __init__(self, user_model: t.Type["User"], role_model: t.Type["Role"]):
         self.user_model = user_model
         self.role_model = role_model
 
-    def _prepare_role_modify_args(self, user, role):
-        if isinstance(user, string_types):
-            user = self.find_user(email=user)
-        if isinstance(role, string_types):
-            role = self.find_role(role)
-        return user, role
+    def _prepare_role_modify_args(
+        self, role: t.Union[str, "Role"]
+    ) -> t.Union["Role", None]:
+        if isinstance(role, str):
+            return self.find_role(role)
+        return role
 
     def _prepare_create_user_args(self, **kwargs):
         kwargs.setdefault("active", True)
@@ -147,62 +153,59 @@ class UserDatastore(object):
             # see if the role exists
             roles[i] = self.find_role(rn)
         kwargs["roles"] = roles
-        if hasattr(self.user_model, "fs_uniquifier"):
-            kwargs.setdefault("fs_uniquifier", uuid.uuid4().hex)
+        kwargs.setdefault("fs_uniquifier", uuid.uuid4().hex)
+        if hasattr(self.user_model, "fs_token_uniquifier"):
+            kwargs.setdefault("fs_token_uniquifier", uuid.uuid4().hex)
+
         return kwargs
 
-    def _is_numeric(self, value):
-        try:
-            int(value)
-        except (TypeError, ValueError):
-            return False
-        return True
-
-    def _is_uuid(self, value):
-        return isinstance(value, uuid.UUID)
-
-    def get_user(self, id_or_email):
-        """Returns a user matching the specified ID or email address."""
-        raise NotImplementedError
-
-    def find_user(self, *args, **kwargs):
+    def find_user(
+        self, case_insensitive: bool = False, **kwargs: t.Any
+    ) -> t.Union["User", None]:
         """Returns a user matching the provided parameters."""
         raise NotImplementedError
 
-    def find_role(self, *args, **kwargs):
+    def find_role(self, role: str) -> t.Union["Role", None]:
         """Returns a role matching the provided name."""
         raise NotImplementedError
 
-    def add_role_to_user(self, user, role):
+    def add_role_to_user(self, user: "User", role: t.Union["Role", str]) -> bool:
         """Adds a role to a user.
 
-        :param user: The user to manipulate. Can be an User object or email
+        :param user: The user to manipulate.
         :param role: The role to add to the user. Can be a Role object or
             string role name
+        :return: True is role was added, False if role already existed.
         """
-        user, role = self._prepare_role_modify_args(user, role)
-        if role not in user.roles:
-            user.roles.append(role)
-            self.put(user)
+        role_obj = self._prepare_role_modify_args(role)
+        if not role_obj:
+            raise ValueError(f"Role: {role} doesn't exist")
+        if role_obj not in user.roles:
+            user.roles.append(role_obj)
+            self.put(user)  # type: ignore
             return True
         return False
 
-    def remove_role_from_user(self, user, role):
+    def remove_role_from_user(self, user: "User", role: t.Union["Role", str]) -> bool:
         """Removes a role from a user.
 
         :param user: The user to manipulate. Can be an User object or email
         :param role: The role to remove from the user. Can be a Role object or
             string role name
+        :return: True if role was removed, False if role doesn't exist or user didn't
+            have role.
         """
         rv = False
-        user, role = self._prepare_role_modify_args(user, role)
-        if role in user.roles:
+        role_obj = self._prepare_role_modify_args(role)
+        if role_obj in user.roles:
             rv = True
-            user.roles.remove(role)
-            self.put(user)
+            user.roles.remove(role_obj)
+            self.put(user)  # type: ignore
         return rv
 
-    def add_permissions_to_role(self, role, permissions):
+    def add_permissions_to_role(
+        self, role: t.Union["Role", str], permissions: t.Union[set, list, str]
+    ) -> bool:
         """Add one or more permissions to role.
 
         :param role: The role to modify. Can be a Role object or
@@ -212,18 +215,20 @@ class UserDatastore(object):
 
         Caller must commit to DB.
 
-        .. versionadded:: 3.4.4
+        .. versionadded:: 4.0.0
         """
 
         rv = False
-        user, role = self._prepare_role_modify_args(None, role)
-        if role:
+        role_obj = self._prepare_role_modify_args(role)
+        if role_obj:
             rv = True
-            role.add_permissions(permissions)
-            self.put(role)
+            role_obj.add_permissions(permissions)
+            self.put(role_obj)  # type: ignore
         return rv
 
-    def remove_permissions_from_role(self, role, permissions):
+    def remove_permissions_from_role(
+        self, role: t.Union["Role", str], permissions: t.Union[set, list, str]
+    ) -> bool:
         """Remove one or more permissions from a role.
 
         :param role: The role to modify. Can be a Role object or
@@ -233,24 +238,24 @@ class UserDatastore(object):
 
         Caller must commit to DB.
 
-        .. versionadded:: 3.4.4
+        .. versionadded:: 4.0.0
         """
 
         rv = False
-        user, role = self._prepare_role_modify_args(None, role)
-        if role:
+        role_obj = self._prepare_role_modify_args(role)
+        if role_obj:
             rv = True
-            role.remove_permissions(permissions)
-            self.put(role)
+            role_obj.remove_permissions(permissions)
+            self.put(role_obj)  # type: ignore
         return rv
 
-    def toggle_active(self, user):
+    def toggle_active(self, user: "User") -> bool:
         """Toggles a user's active status. Always returns True."""
         user.active = not user.active
-        self.put(user)
+        self.put(user)  # type: ignore
         return True
 
-    def deactivate_user(self, user):
+    def deactivate_user(self, user: "User") -> bool:
         """Deactivates a specified user. Returns `True` if a change was made.
 
         This will immediately disallow access to all endpoints that require
@@ -261,42 +266,59 @@ class UserDatastore(object):
         """
         if user.active:
             user.active = False
-            self.put(user)
+            self.put(user)  # type: ignore
             return True
         return False
 
-    def activate_user(self, user):
+    def activate_user(self, user: "User") -> bool:
         """Activates a specified user. Returns `True` if a change was made.
 
         :param user: The user to activate
         """
         if not user.active:
             user.active = True
-            self.put(user)
+            self.put(user)  # type: ignore
             return True
         return False
 
-    def set_uniquifier(self, user, uniquifier=None):
-        """ Set user's authentication token uniquifier.
+    def set_uniquifier(
+        self, user: "User", uniquifier: t.Union[str, None] = None
+    ) -> None:
+        """Set user's Flask-Security identity key.
         This will immediately render outstanding auth tokens,
         session cookies and remember cookies invalid.
 
         :param user: User to modify
         :param uniquifier: Unique value - if none then uuid.uuid4().hex is used
 
-        This method is a no-op if the user model doesn't contain the attribute
-        ``fs_uniquifier``
-
         .. versionadded:: 3.3.0
         """
-        if not hasattr(user, "fs_uniquifier"):
-            return
         if not uniquifier:
             uniquifier = uuid.uuid4().hex
         user.fs_uniquifier = uniquifier
-        self.put(user)
+        self.put(user)  # type: ignore
 
-    def create_role(self, **kwargs):
+    def set_token_uniquifier(
+        self, user: "User", uniquifier: t.Union[str, None] = None
+    ) -> None:
+        """Set user's auth token identity key.
+        This will immediately render outstanding auth tokens invalid.
+
+        :param user: User to modify
+        :param uniquifier: Unique value - if none then uuid.uuid4().hex is used
+
+        This method is a no-op if the user model doesn't contain the attribute
+        ``fs_token_uniquifier``
+
+        .. versionadded:: 4.0.0
+        """
+        if not uniquifier:
+            uniquifier = uuid.uuid4().hex
+        if hasattr(user, "fs_token_uniquifier"):
+            user.fs_token_uniquifier = uniquifier
+            self.put(user)  # type: ignore
+
+    def create_role(self, **kwargs: t.Any) -> "Role":
         """
         Creates and returns a new role from the given parameters.
         Supported params (depending on RoleModel):
@@ -316,22 +338,21 @@ class UserDatastore(object):
             perms = kwargs["permissions"]
             if isinstance(perms, list) or isinstance(perms, set):
                 perms = ",".join(perms)
-            elif isinstance(perms, string_types):
+            elif isinstance(perms, str):
                 # squash spaces.
-                perms = ",".join([p.strip() for p in perms.split(",")])
+                perms = ",".join(p.strip() for p in perms.split(","))
             kwargs["permissions"] = perms
 
         role = self.role_model(**kwargs)
-        return self.put(role)
+        return self.put(role)  # type: ignore
 
-    def find_or_create_role(self, name, **kwargs):
+    def find_or_create_role(self, name: str, **kwargs: t.Any) -> "Role":
         """Returns a role matching the given name or creates it with any
         additionally provided parameters.
         """
-        kwargs["name"] = name
-        return self.find_role(name) or self.create_role(**kwargs)
+        return self.find_role(name) or self.create_role(name=name, **kwargs)
 
-    def create_user(self, **kwargs):
+    def create_user(self, **kwargs: t.Any) -> "User":
         """Creates and returns a new user from the given parameters.
 
         :kwparam email: required.
@@ -339,60 +360,88 @@ class UserDatastore(object):
         :kwparam roles: list of roles to be added to user.
             Can be Role objects or strings
 
+        Any other element of the User data model may be supplied as well.
+
+        .. note::
+            No normalization is done on email - it is assumed the caller has already
+            done that.
+
+            Best practice is::
+
+                try:
+                    enorm = app.security._mail_util.validate(email)
+                except ValueError:
+
+        .. note::
+            The roles kwparam is modified as part of the call - it will, if necessary
+            be converted from names to role instances.
+
         .. danger::
            Be aware that whatever `password` is passed in will
            be stored directly in the DB. Do NOT pass in a plaintext password!
            Best practice is to pass in ``hash_password(plaintext_password)``.
 
-           Furthermore, no validation is done on the password (e.g for minimum length).
-           Best practice is to call
-           ``app.security._password_validator(plaintext_password, True)``
-           and look for a ``None`` return meaning the password conforms to the
-           configured validations.
+           Furthermore, no validation nor normalization is done on the password
+           (e.g for minimum length).
+
+           Best practice is::
+
+            pbad, pnorm = app.security._password_util.validate(password, True)
+
+           Look for `pbad` being None. Pass the normalized password `pnorm` to this
+           method.
 
         The new user's ``active`` property will be set to ``True``
-        unless explicitly set to ``False`` in `kwargs`.
+        unless explicitly set to ``False`` in `kwargs` (e.g. active = False)
         """
         kwargs = self._prepare_create_user_args(**kwargs)
         user = self.user_model(**kwargs)
-        return self.put(user)
+        return self.put(user)  # type: ignore
 
-    def delete_user(self, user):
+    def delete_user(self, user: "User") -> None:
         """Deletes the specified user.
 
         :param user: The user to delete
         """
-        self.delete(user)
+        self.delete(user)  # type: ignore
 
-    def reset_user_access(self, user):
+    def reset_user_access(self, user: "User") -> None:
         """
         Use this method to reset user authentication methods in the case of compromise.
         This will:
 
             * reset fs_uniquifier - which causes session cookie, remember cookie, auth
               tokens to be unusable
+            * reset fs_token_uniquifier (if present) - cause auth tokens to be unusable
             * remove all unified signin TOTP secrets so those can't be used
             * remove all two-factor secrets so those can't be used
 
-        Note that if using unified sign in and allow 'email' as a way to receive a code
+        Note that if using unified sign in and allow 'email' as a way to receive a code;
         if the email is compromised - login is still possible. To handle this - it
         is better to deactivate the user.
 
         Note - this method isn't used directly by Flask-Security - it is provided
-        as a helper for an applications administrative needs.
+        as a helper for an application's administrative needs.
 
         Remember to call commit on DB if needed.
 
         .. versionadded:: 3.4.1
         """
         self.set_uniquifier(user)
+        self.set_token_uniquifier(user)
         if hasattr(user, "us_totp_secrets"):
             self.us_reset(user)
         if hasattr(user, "tf_primary_method"):
             self.tf_reset(user)
 
-    def tf_set(self, user, primary_method, totp_secret=None, phone=None):
-        """ Set two-factor info into user record.
+    def tf_set(
+        self,
+        user: "User",
+        primary_method: str,
+        totp_secret: t.Optional[str] = None,
+        phone: t.Optional[str] = None,
+    ) -> None:
+        """Set two-factor info into user record.
         This carefully only changes things if different.
 
         If totp_secret isn't provided - existing one won't be changed.
@@ -417,20 +466,20 @@ class UserDatastore(object):
             user.tf_phone_number = phone
             changed = True
         if changed:
-            self.put(user)
+            self.put(user)  # type: ignore
 
-    def tf_reset(self, user):
-        """ Disable two-factor auth for user
+    def tf_reset(self, user: "User") -> None:
+        """Disable two-factor auth for user
 
         .. versionadded: 3.4.1
         """
         user.tf_primary_method = None
         user.tf_totp_secret = None
         user.tf_phone_number = None
-        self.put(user)
+        self.put(user)  # type: ignore
 
-    def us_get_totp_secrets(self, user):
-        """ Return totp secrets.
+    def us_get_totp_secrets(self, user: "User") -> t.Dict[str, str]:
+        """Return totp secrets.
         These are json encoded in the DB.
 
         Returns a dict with methods as keys and secrets as values.
@@ -441,17 +490,25 @@ class UserDatastore(object):
             return {}
         return json.loads(user.us_totp_secrets)
 
-    def us_put_totp_secrets(self, user, secrets):
-        """ Save secrets. Assume to be a dict (or None)
+    def us_put_totp_secrets(
+        self, user: "User", secrets: t.Optional[t.Dict[str, str]]
+    ) -> None:
+        """Save secrets. Assume to be a dict (or None)
         with keys as methods, and values as (encrypted) secrets.
 
         .. versionadded:: 3.4.0
         """
         user.us_totp_secrets = json.dumps(secrets) if secrets else None
-        self.put(user)
+        self.put(user)  # type: ignore
 
-    def us_set(self, user, method, totp_secret=None, phone=None):
-        """ Set unified sign in info into user record.
+    def us_set(
+        self,
+        user: "User",
+        method: str,
+        totp_secret: t.Optional[str] = None,
+        phone: t.Optional[str] = None,
+    ) -> None:
+        """Set unified sign in info into user record.
 
         If totp_secret isn't provided - existing one won't be changed.
         If phone isn't provided, the existing phone number won't be changed.
@@ -470,10 +527,10 @@ class UserDatastore(object):
             self.us_put_totp_secrets(user, totp_secrets)
         if phone and user.us_phone_number != phone:
             user.us_phone_number = phone
-            self.put(user)
+            self.put(user)  # type: ignore
 
-    def us_reset(self, user):
-        """ Disable unified sign in for user.
+    def us_reset(self, user: "User") -> None:
+        """Disable unified sign in for user.
         Be aware that if "email" is an allowed way to receive codes, they
         will still work (as totp secrets are generated on the fly).
         This will disable authenticator app and SMS.
@@ -481,91 +538,76 @@ class UserDatastore(object):
         .. versionadded: 3.4.1
         """
         user.us_totp_secrets = None
-        user.us_phone_number = None
-        self.put(user)
+        self.put(user)  # type: ignore
 
 
 class SQLAlchemyUserDatastore(SQLAlchemyDatastore, UserDatastore):
-    """A SQLAlchemy datastore implementation for Flask-Security that assumes the
-    use of the Flask-SQLAlchemy extension.
+    """A UserDatastore implementation that assumes the
+    use of
+    `Flask-SQLAlchemy <https://pypi.python.org/pypi/flask-sqlalchemy/>`_
+    for datastore transactions.
+
+    :param db:
+    :param user_model: See :ref:`Models <models_topic>`.
+    :param role_model: See :ref:`Models <models_topic>`.
     """
 
-    def __init__(self, db, user_model, role_model):
+    def __init__(
+        self,
+        db: "flask_sqlalchemy.SQLAlchemy",
+        user_model: t.Type["User"],
+        role_model: t.Type["Role"],
+    ):
         SQLAlchemyDatastore.__init__(self, db)
         UserDatastore.__init__(self, user_model, role_model)
 
-    def get_user(self, identifier):
+    def find_user(
+        self, case_insensitive: bool = False, **kwargs: t.Any
+    ) -> t.Union["User", None]:
         from sqlalchemy import func as alchemyFn
-        from sqlalchemy import inspect
-        from sqlalchemy.sql import sqltypes
-        from sqlalchemy.dialects.postgresql import UUID as PSQL_UUID
 
-        user_model_query = self.user_model.query
-        if config_value("JOIN_USER_ROLES") and hasattr(self.user_model, "roles"):
-            from sqlalchemy.orm import joinedload
-
-            user_model_query = user_model_query.options(joinedload("roles"))
-
-        # To support both numeric, string, and UUID primary keys, and support
-        # calling this routine with either a numeric value or a string or a UUID
-        # we need to make sure the types basically match.
-        # psycopg2 for example will complain if we attempt to 'get' a
-        # numeric primary key with a string value.
-        # TODO: other datastores don't support this - they assume the only
-        # PK is user.id. That makes things easier but for backwards compat...
-        ins = inspect(self.user_model)
-        pk_type = ins.primary_key[0].type
-        pk_isnumeric = isinstance(pk_type, sqltypes.Integer)
-        pk_isuuid = isinstance(pk_type, PSQL_UUID)
-        # Are they the same or NOT numeric nor UUID
-        if (
-            (pk_isnumeric and self._is_numeric(identifier))
-            or (pk_isuuid and self._is_uuid(identifier))
-            or (not pk_isnumeric and not pk_isuuid)
-        ):
-            rv = self.user_model.query.get(identifier)
-            if rv is not None:
-                return rv
-
-        # Not PK - iterate through other attributes and look for 'identifier'
-        for attr in get_identity_attributes():
-            column = getattr(self.user_model, attr)
-            attr_isnumeric = isinstance(column.type, sqltypes.Integer)
-
-            query = None
-            if attr_isnumeric and self._is_numeric(identifier):
-                query = column == identifier
-            elif not attr_isnumeric and not self._is_numeric(identifier):
-                # Look for exact case-insensitive match - 'ilike' honors
-                # wild cards which isn't what we want.
-                query = alchemyFn.lower(column) == alchemyFn.lower(identifier)
-            if query is not None:
-                rv = user_model_query.filter(query).first()
-                if rv is not None:
-                    return rv
-
-    def find_user(self, **kwargs):
         query = self.user_model.query
         if config_value("JOIN_USER_ROLES") and hasattr(self.user_model, "roles"):
             from sqlalchemy.orm import joinedload
 
             query = query.options(joinedload("roles"))
 
-        return query.filter_by(**kwargs).first()
+        if case_insensitive:
+            # While it is of course possible to pass in multiple keys to filter on
+            # that isn't the normal use case. If caller asks for case_insensitive
+            # AND gives multiple keys - throw an error.
+            if len(kwargs) > 1:
+                raise ValueError("Case insensitive option only supports single key")
+            attr, identifier = kwargs.popitem()
+            subquery = alchemyFn.lower(
+                getattr(self.user_model, attr)
+            ) == alchemyFn.lower(identifier)
+            return query.filter(subquery).first()
+        else:
+            return query.filter_by(**kwargs).first()
 
-    def find_role(self, role):
-        return self.role_model.query.filter_by(name=role).first()
+    def find_role(self, role: str) -> t.Union["Role", None]:
+        return self.role_model.query.filter_by(name=role).first()  # type: ignore
 
 
 class SQLAlchemySessionUserDatastore(SQLAlchemyUserDatastore, SQLAlchemyDatastore):
-    """A SQLAlchemy datastore implementation for Flask-Security that assumes the
-    use of the flask_sqlalchemy_session extension.
+    """A UserDatastore implementation that directly uses
+    `SQLAlchemy's <https://docs.sqlalchemy.org/en/14/orm/session_basics.html>`_
+    session API.
+
+    :param session:
+    :param user_model: See :ref:`Models <models_topic>`.
+    :param role_model: See :ref:`Models <models_topic>`.
     """
 
-    def __init__(self, session, user_model, role_model):
-        class PretendFlaskSQLAlchemyDb(object):
-            """ This is a pretend db object, so we can just pass in a session.
-            """
+    def __init__(
+        self,
+        session: "sqlalchemy.orm.scoping.scoped_session",
+        user_model: t.Type["User"],
+        role_model: t.Type["Role"],
+    ):
+        class PretendFlaskSQLAlchemyDb:
+            """This is a pretend db object, so we can just pass in a session."""
 
             def __init__(self, session):
                 self.session = session
@@ -575,51 +617,47 @@ class SQLAlchemySessionUserDatastore(SQLAlchemyUserDatastore, SQLAlchemyDatastor
         )
 
     def commit(self):
-        super(SQLAlchemySessionUserDatastore, self).commit()
+        super().commit()
 
 
 class MongoEngineUserDatastore(MongoEngineDatastore, UserDatastore):
-    """A MongoEngine datastore implementation for Flask-Security that assumes
-    the use of the Flask-MongoEngine extension.
+    """A UserDatastore implementation that assumes the
+    use of
+    `Flask-MongoEngine <https://pypi.python.org/pypi/flask-mongoengine/>`_
+    for datastore transactions.
+
+    :param db:
+    :param user_model: See :ref:`Models <models_topic>`.
+    :param role_model: See :ref:`Models <models_topic>`.
     """
 
-    def __init__(self, db, user_model, role_model):
+    def __init__(
+        self,
+        db: "flask_mongoengine.MongoEngine",
+        user_model: t.Type["User"],
+        role_model: t.Type["Role"],
+    ):
         MongoEngineDatastore.__init__(self, db)
         UserDatastore.__init__(self, user_model, role_model)
 
-    def get_user(self, identifier):
-        from mongoengine import ValidationError
-
-        try:
-            return self.user_model.objects(id=identifier).first()
-        except (ValidationError, ValueError):
-            pass
-
-        is_numeric = self._is_numeric(identifier)
-
-        for attr in get_identity_attributes():
-            query_key = attr if is_numeric else "%s__iexact" % attr
-            query = {query_key: identifier}
-            try:
-                rv = self.user_model.objects(**query).first()
-                if rv is not None:
-                    return rv
-            except (ValidationError, ValueError):
-                # This can happen if identifier is a string but attribute is
-                # an int.
-                pass
-
-    def find_user(self, **kwargs):
-        try:
-            from mongoengine.queryset import Q, QCombination
-        except ImportError:
-            from mongoengine.queryset.visitor import Q, QCombination
+    def find_user(self, case_insensitive=False, **kwargs):
+        from mongoengine.queryset.visitor import Q, QCombination
         from mongoengine.errors import ValidationError
 
-        queries = map(lambda i: Q(**{i[0]: i[1]}), kwargs.items())
-        query = QCombination(QCombination.AND, queries)
         try:
-            return self.user_model.objects(query).first()
+            if case_insensitive:
+                # While it is of course possible to pass in multiple keys to filter on
+                # that isn't the normal use case. If caller asks for case_insensitive
+                # AND gives multiple keys - throw an error.
+                if len(kwargs) > 1:
+                    raise ValueError("Case insensitive option only supports single key")
+                attr, identifier = kwargs.popitem()
+                query = {f"{attr}__iexact": identifier}
+                return self.user_model.objects(**query).first()
+            else:
+                queries = map(lambda i: Q(**{i[0]: i[1]}), kwargs.items())
+                query = QCombination(QCombination.AND, queries)
+                return self.user_model.objects(query).first()
         except ValidationError:  # pragma: no cover
             return None
 
@@ -628,47 +666,46 @@ class MongoEngineUserDatastore(MongoEngineDatastore, UserDatastore):
 
 
 class PeeweeUserDatastore(PeeweeDatastore, UserDatastore):
-    """A PeeweeD datastore implementation for Flask-Security that assumes the
-    use of Peewee Flask utils.
+    """A UserDatastore implementation that assumes the
+    use of
+    `Peewee Flask utils \
+       <https://docs.peewee-orm.com/en/latest/peewee/playhouse.html#flask-utils>`_
+    for datastore transactions.
 
-    :param user_model: A user model class definition
-    :param role_model: A role model class definition
-    :param role_link: A model implementing the many-to-many user-role relation
+    :param db:
+    :param user_model: See :ref:`Models <models_topic>`.
+    :param role_model: See :ref:`Models <models_topic>`.
+    :param role_link:
     """
 
     def __init__(self, db, user_model, role_model, role_link):
+        """
+        :param db:
+        :param user_model: A user model class definition
+        :param role_model: A role model class definition
+        :param role_link: A model implementing the many-to-many user-role relation
+        """
         PeeweeDatastore.__init__(self, db)
         UserDatastore.__init__(self, user_model, role_model)
         self.UserRole = role_link
 
-    def get_user(self, identifier):
+    def find_user(self, case_insensitive=False, **kwargs):
         from peewee import fn as peeweeFn
-        from peewee import IntegerField
 
-        # For peewee we only (currently) support numeric primary keys.
-        if self._is_numeric(identifier):
-            try:
-                return self.user_model.get(self.user_model.id == identifier)
-            except (self.user_model.DoesNotExist, ValueError):
-                pass
-
-        for attr in get_identity_attributes():
-            # Read above (SQLAlchemy store) for why we are checking types.
-            column = getattr(self.user_model, attr)
-            attr_isnumeric = isinstance(column, IntegerField)
-            try:
-                if attr_isnumeric and self._is_numeric(identifier):
-                    return self.user_model.get(column == identifier)
-                elif not attr_isnumeric and not self._is_numeric(identifier):
-                    return self.user_model.get(
-                        peeweeFn.Lower(column) == peeweeFn.Lower(identifier)
-                    )
-            except (self.user_model.DoesNotExist, ValueError):
-                pass
-
-    def find_user(self, **kwargs):
         try:
-            return self.user_model.filter(**kwargs).get()
+            if case_insensitive:
+                # While it is of course possible to pass in multiple keys to filter on
+                # that isn't the normal use case. If caller asks for case_insensitive
+                # AND gives multiple keys - throw an error.
+                if len(kwargs) > 1:
+                    raise ValueError("Case insensitive option only supports single key")
+                attr, identifier = kwargs.popitem()
+                return self.user_model.get(
+                    peeweeFn.lower(getattr(self.user_model, attr))
+                    == peeweeFn.lower(identifier)
+                )
+            else:
+                return self.user_model.filter(**kwargs).get()
         except self.user_model.DoesNotExist:
             return None
 
@@ -694,7 +731,7 @@ class PeeweeUserDatastore(PeeweeDatastore, UserDatastore):
         :param user: The user to manipulate
         :param role: The role to add to the user
         """
-        user, role = self._prepare_role_modify_args(user, role)
+        role = self._prepare_role_modify_args(role)
         result = self.UserRole.select().where(
             self.UserRole.user == user.id, self.UserRole.role == role.id
         )
@@ -710,7 +747,7 @@ class PeeweeUserDatastore(PeeweeDatastore, UserDatastore):
         :param user: The user to manipulate
         :param role: The role to remove from the user
         """
-        user, role = self._prepare_role_modify_args(user, role)
+        role = self._prepare_role_modify_args(role)
         result = self.UserRole.select().where(
             self.UserRole.user == user, self.UserRole.role == role
         )
@@ -725,10 +762,17 @@ class PeeweeUserDatastore(PeeweeDatastore, UserDatastore):
 
 
 class PonyUserDatastore(PonyDatastore, UserDatastore):
-    """A Pony ORM datastore implementation for Flask-Security.
+    """A UserDatastore implementation that assumes the
+    use of
+    `PonyORM <https://pypi.python.org/pypi/pony/>`_
+    for datastore transactions.
 
     Code primarily from https://github.com/ET-CS but taken over after
     being abandoned.
+
+    :param db:
+    :param user_model: See :ref:`Models <models_topic>`.
+    :param role_model: See :ref:`Models <models_topic>`.
     """
 
     def __init__(self, db, user_model, role_model):
@@ -736,26 +780,15 @@ class PonyUserDatastore(PonyDatastore, UserDatastore):
         UserDatastore.__init__(self, user_model, role_model)
 
     @with_pony_session
-    def get_user(self, identifier):
-        from pony.orm.core import ObjectNotFound
+    def find_user(self, case_insensitive=False, **kwargs):
+        if case_insensitive:
+            # While it is of course possible to pass in multiple keys to filter on
+            # that isn't the normal use case. If caller asks for case_insensitive
+            # AND gives multiple keys - throw an error.
+            if len(kwargs) > 1:
+                raise ValueError("Case insensitive option only supports single key")
+            # TODO - implement case insensitive look ups.
 
-        try:
-            return self.user_model[identifier]
-        except (ObjectNotFound, ValueError):
-            pass
-
-        for attr in get_identity_attributes():
-            # this is a nightmare, tl;dr we need to get the thing that
-            # corresponds to email (usually)
-            try:
-                user = self.user_model.get(**{attr: identifier})
-                if user is not None:
-                    return user
-            except (TypeError, ValueError):
-                pass
-
-    @with_pony_session
-    def find_user(self, **kwargs):
         return self.user_model.get(**kwargs)
 
     @with_pony_session
@@ -764,12 +797,58 @@ class PonyUserDatastore(PonyDatastore, UserDatastore):
 
     @with_pony_session
     def add_role_to_user(self, *args, **kwargs):
-        return super(PonyUserDatastore, self).add_role_to_user(*args, **kwargs)
+        return super().add_role_to_user(*args, **kwargs)
 
     @with_pony_session
     def create_user(self, **kwargs):
-        return super(PonyUserDatastore, self).create_user(**kwargs)
+        return super().create_user(**kwargs)
 
     @with_pony_session
     def create_role(self, **kwargs):
-        return super(PonyUserDatastore, self).create_role(**kwargs)
+        return super().create_role(**kwargs)
+
+
+if t.TYPE_CHECKING:  # pragma: no cover
+    # Normally - the application creates the Models and glues them together
+    # For typing we do that here since we don't know which DB interface they
+    # will pick.
+    from .core import UserMixin, RoleMixin
+
+    class CanonicalUserDatastore(Datastore, UserDatastore):
+        pass
+
+    class User(UserMixin):
+        id: int
+        email: str
+        username: t.Optional[str]
+        password: str
+        active: bool
+        fs_uniquifier: str
+        fs_token_uniquifier: str
+        confirmed_at: datetime.datetime
+        last_login_at: datetime.datetime
+        current_login_at: datetime.datetime
+        last_login_ip: t.Optional[str]
+        current_login_ip: t.Optional[str]
+        login_count: int
+        tf_primary_method: t.Optional[str]
+        tf_totp_secret: t.Optional[str]
+        tf_phone_number: t.Optional[str]
+        us_phone_number: t.Optional[str]
+        us_totp_secrets: t.Optional[t.Union[str, bytes]]
+        create_datetime: datetime.datetime
+        update_datetime: datetime.datetime
+        roles: t.List["Role"]
+
+        def __init__(self, **kwargs):
+            ...
+
+    class Role(RoleMixin):
+        id: int
+        name: str
+        description: t.Optional[str]
+        permissions: t.Optional[t.Union[str, set, list]]
+        update_datetime: datetime.datetime
+
+        def __init__(self, **kwargs):
+            ...
