@@ -1,4 +1,3 @@
-# -*- coding: utf-8 -*-
 """
     flask_security.decorators
     ~~~~~~~~~~~~~~~~~~~~~~~~~
@@ -6,13 +5,14 @@
     Flask-Security decorators module
 
     :copyright: (c) 2012-2019 by Matt Wright.
-    :copyright: (c) 2019-2020 by J. Christopher Wagner (jwag).
+    :copyright: (c) 2019-2021 by J. Christopher Wagner (jwag).
     :license: MIT, see LICENSE for more details.
 """
 
 from collections import namedtuple
 import datetime
 from functools import wraps
+import typing as t
 
 from flask import Response, _request_ctx_stack, abort, current_app, g, redirect, request
 from flask_login import current_user, login_required  # noqa: F401
@@ -21,25 +21,26 @@ from flask_wtf.csrf import CSRFError
 from werkzeug.local import LocalProxy
 from werkzeug.routing import BuildError
 
+from .proxies import _security, DecoratedView
 from .utils import (
     FsPermNeed,
     config_value,
     do_flash,
+    find_user,
     get_message,
     get_url,
     check_and_update_authn_fresh,
     json_error_response,
+    set_request_attr,
 )
 
 # Convenient references
-_security = LocalProxy(lambda: current_app.extensions["security"])
-
 _csrf = LocalProxy(lambda: current_app.extensions["csrf"])
 
 BasicAuth = namedtuple("BasicAuth", "username, password")
 
 # NOTE: this is here for backwards compatibility, it is deprecated and
-# to be removed in 4.0
+# could be removed any time!
 _default_unauthenticated_html = """
     <h1>Unauthorized</h1>
     <p>The server could not verify that you are authorized to access the URL
@@ -61,7 +62,7 @@ def _get_unauthorized_response(text=None, headers=None):  # pragma: no cover
 
 
 def default_unauthn_handler(mechanisms, headers=None):
-    """ Default callback for failures to authenticate
+    """Default callback for failures to authenticate
 
     If caller wants JSON - return 401.
     If caller wants BasicAuth - return 401 (the WWW-Authenticate header is set).
@@ -75,9 +76,6 @@ def default_unauthn_handler(mechanisms, headers=None):
     if config_value("BACKWARDS_COMPAT_UNAUTHN"):
         return _get_unauthenticated_response(headers=headers)
     if _security._want_json(request):
-        # Remove WWW-Authenticate from headers for JSON responses since
-        # browsers will intercept that and pop up their own login form.
-        headers.pop("WWW-Authenticate", None)
         payload = json_error_response(errors=msg)
         return _security._render_json(payload, 401, headers, None)
 
@@ -90,7 +88,7 @@ def default_unauthn_handler(mechanisms, headers=None):
 
 
 def default_reauthn_handler(within, grace):
-    """ Default callback for 'freshness' related authn failures.
+    """Default callback for 'freshness' related authn failures.
 
     If caller wants JSON - return 401
     Otherwise - assume caller is html and redirect if possible to configured view.
@@ -156,7 +154,7 @@ def _check_token():
     if user and user.is_authenticated:
         app = current_app._get_current_object()
         _request_ctx_stack.top.user = user
-        identity_changed.send(app, identity=Identity(user.id))
+        identity_changed.send(app, identity=Identity(user.fs_uniquifier))
         return True
 
     return False
@@ -166,20 +164,22 @@ def _check_http_auth():
     auth = request.authorization or BasicAuth(username=None, password=None)
     if not auth.username:
         return False
-    user = _security.datastore.get_user(auth.username)
+    user = find_user(auth.username)
+    if user and not user.active:
+        return False
 
     if user and user.verify_and_update_password(auth.password):
         _security.datastore.commit()
         app = current_app._get_current_object()
         _request_ctx_stack.top.user = user
-        identity_changed.send(app, identity=Identity(user.id))
+        identity_changed.send(app, identity=Identity(user.fs_uniquifier))
         return True
 
     return False
 
 
-def handle_csrf(method):
-    """ Invoke CSRF protection based on authentication method.
+def handle_csrf(method: t.Optional[str]) -> None:
+    """Invoke CSRF protection based on authentication method.
 
     Usually this is called as part of a decorator, but if that isn't
     appropriate, endpoint code can call this directly.
@@ -210,12 +210,12 @@ def handle_csrf(method):
 
     if config_value("CSRF_PROTECT_MECHANISMS"):
         if method in config_value("CSRF_PROTECT_MECHANISMS"):
-            _csrf.protect()
+            _csrf.protect()  # type: ignore
         else:
             _request_ctx_stack.top.fs_ignore_csrf = True
 
 
-def http_auth_required(realm):
+def http_auth_required(realm: t.Any) -> DecoratedView:
     """Decorator that protects endpoints using Basic HTTP authentication.
 
     :param realm: optional realm name
@@ -231,6 +231,7 @@ def http_auth_required(realm):
         def wrapper(*args, **kwargs):
             if _check_http_auth():
                 handle_csrf("basic")
+                set_request_attr("fs_authn_via", "basic")
                 return fn(*args, **kwargs)
             if _security._unauthorized_callback:
                 return _security._unauthorized_callback()
@@ -246,7 +247,7 @@ def http_auth_required(realm):
     return decorator
 
 
-def auth_token_required(fn):
+def auth_token_required(fn: DecoratedView) -> DecoratedView:
     """Decorator that protects endpoints using token authentication. The token
     should be added to the request by the client by using a query string
     variable with a name equal to the configuration value of
@@ -260,16 +261,21 @@ def auth_token_required(fn):
     def decorated(*args, **kwargs):
         if _check_token():
             handle_csrf("token")
+            set_request_attr("fs_authn_via", "token")
             return fn(*args, **kwargs)
         if _security._unauthorized_callback:
             return _security._unauthorized_callback()
         else:
             return _security._unauthn_handler(["token"])
 
-    return decorated
+    return t.cast(DecoratedView, decorated)
 
 
-def auth_required(*auth_methods, **kwargs):
+def auth_required(
+    *auth_methods: t.Union[str, t.Callable[[], t.List[str]], None],
+    within: t.Union[int, float, t.Callable[[], datetime.timedelta]] = -1,
+    grace: t.Optional[t.Union[int, float, t.Callable[[], datetime.timedelta]]] = None
+) -> DecoratedView:
     """
     Decorator that protects endpoints through multiple mechanisms
     Example::
@@ -280,8 +286,10 @@ def auth_required(*auth_methods, **kwargs):
             return 'Dashboard'
 
     :param auth_methods: Specified mechanisms (token, basic, session). If not specified
-        then all current available mechanisms (except "basic") will be tried.
-    :kwparam within: Add 'freshness' check to authentication. Is either an int
+        then all current available mechanisms (except "basic") will be tried. A callable
+        can also be passed (useful if you need app/request context). The callable
+        must return a list.
+    :param within: Add 'freshness' check to authentication. Is either an int
         specifying # of minutes, or a callable that returns a timedelta. For timedeltas,
         timedelta.total_seconds() is used for the calculations:
 
@@ -293,7 +301,7 @@ def auth_required(*auth_methods, **kwargs):
 
         Note that Basic Auth, by definition, is always 'fresh' and will never result in
         a redirect/error.
-    :kwparam grace: Add a grace period for freshness checks. As above, either an int
+    :param grace: Add a grace period for freshness checks. As above, either an int
         or a callable returning a timedelta. If not specified then
         :py:data:`SECURITY_FRESHNESS_GRACE_PERIOD` is used. The grace period allows
         callers to complete the required operations w/o being prompted again.
@@ -308,6 +316,9 @@ def auth_required(*auth_methods, **kwargs):
     On authentication failure `.Security.unauthorized_callback` (deprecated)
     or :meth:`.Security.unauthn_handler` will be called.
 
+    As a side effect, upon successful authentication, the request global
+     ``fs_authn_via`` will be set to the method ("basic", "token", "session")
+
     .. note::
         If "basic" is specified in addition to other methods, then if authentication
         fails, a 401 with the "WWW-Authenticate" header will be returned - rather than
@@ -324,6 +335,9 @@ def auth_required(*auth_methods, **kwargs):
     .. versionchanged:: 3.4.4
         If ``auth_methods`` isn't specified try all mechanisms EXCEPT ``basic``.
 
+    .. versionchanged:: 4.0.0
+        auth_methods can be passed as a callable.
+
     """
 
     login_mechanisms = {
@@ -332,21 +346,16 @@ def auth_required(*auth_methods, **kwargs):
         "basic": lambda: _check_http_auth(),
     }
     mechanisms_order = ["token", "session", "basic"]
-    if not auth_methods:
-        auth_methods = {"session", "token"}
-    else:
-        auth_methods = [am for am in auth_methods]
 
     def wrapper(fn):
         @wraps(fn)
-        def decorated_view(*args, **dkwargs):
-            # 2.7 doesn't support keyword args after *args....
-            within = kwargs.get("within", -1)
+        def decorated_view(
+            *args, auth_methods=auth_methods, within=within, grace=grace, **kwargs
+        ):
             if callable(within):
                 within = within()
             else:
                 within = datetime.timedelta(minutes=within)
-            grace = kwargs.get("grace", None)
             if grace is None:
                 grace = config_value("FRESHNESS_GRACE_PERIOD")
             elif callable(grace):
@@ -354,37 +363,48 @@ def auth_required(*auth_methods, **kwargs):
             else:
                 grace = datetime.timedelta(minutes=grace)
 
+            if not auth_methods:
+                ams = {"session", "token"}
+            else:
+                ams = []
+                for am in auth_methods:
+                    if callable(am):
+                        ams.extend(am())
+                    else:
+                        ams.append(am)
+
             h = {}
-            if "basic" in auth_methods:
+            if "basic" in ams:
                 r = _security.default_http_auth_realm
                 h["WWW-Authenticate"] = 'Basic realm="%s"' % r
             mechanisms = [
                 (method, login_mechanisms.get(method))
                 for method in mechanisms_order
-                if method in auth_methods
+                if method in ams
             ]
             for method, mechanism in mechanisms:
                 if mechanism and mechanism():
                     # successfully authenticated. Basic auth is by definition 'fresh'.
                     # Note that using token auth is ok - but caller still has to pass
                     # in a session cookie...
-                    if method != "basic" and not check_and_update_authn_fresh(
-                        within, grace
-                    ):
+                    if not check_and_update_authn_fresh(within, grace, method):
                         return _security._reauthn_handler(within, grace)
                     handle_csrf(method)
-                    return fn(*args, **dkwargs)
+                    set_request_attr("fs_authn_via", method)
+                    return fn(*args, **kwargs)
             if _security._unauthorized_callback:
                 return _security._unauthorized_callback()
             else:
-                return _security._unauthn_handler(auth_methods, headers=h)
+                return _security._unauthn_handler(ams, headers=h)
 
         return decorated_view
 
     return wrapper
 
 
-def unauth_csrf(fall_through=False):
+def unauth_csrf(
+    fall_through: bool = False,
+) -> DecoratedView:
     """Decorator for endpoints that don't need authentication
     but do want CSRF checks (available via Header rather than just form).
     This is required when setting *WTF_CSRF_CHECK_DEFAULT* = **False** since in that
@@ -438,7 +458,7 @@ def unauth_csrf(fall_through=False):
     return wrapper
 
 
-def roles_required(*roles):
+def roles_required(*roles: str) -> DecoratedView:
     """Decorator which specifies that a user must have all the specified roles.
     Example::
 
@@ -470,7 +490,7 @@ def roles_required(*roles):
     return wrapper
 
 
-def roles_accepted(*roles):
+def roles_accepted(*roles: str) -> DecoratedView:
     """Decorator which specifies that a user must have at least one of the
     specified roles. Example::
 
@@ -488,7 +508,7 @@ def roles_accepted(*roles):
     def wrapper(fn):
         @wraps(fn)
         def decorated_view(*args, **kwargs):
-            perm = Permission(*[RoleNeed(role) for role in roles])
+            perm = Permission(*(RoleNeed(role) for role in roles))
             if perm.can():
                 return fn(*args, **kwargs)
             if _security._unauthorized_callback:
@@ -501,7 +521,7 @@ def roles_accepted(*roles):
     return wrapper
 
 
-def permissions_required(*fsperms):
+def permissions_required(*fsperms: str) -> DecoratedView:
     """Decorator which specifies that a user must have all the specified permissions.
     Example::
 
@@ -540,7 +560,7 @@ def permissions_required(*fsperms):
     return wrapper
 
 
-def permissions_accepted(*fsperms):
+def permissions_accepted(*fsperms: str) -> DecoratedView:
     """Decorator which specifies that a user must have at least one of the
     specified permissions. Example::
 
@@ -562,7 +582,7 @@ def permissions_accepted(*fsperms):
     def wrapper(fn):
         @wraps(fn)
         def decorated_view(*args, **kwargs):
-            perm = Permission(*[FsPermNeed(fsperm) for fsperm in fsperms])
+            perm = Permission(*(FsPermNeed(fsperm) for fsperm in fsperms))
             if perm.can():
                 return fn(*args, **kwargs)
             if _security._unauthorized_callback:
@@ -575,7 +595,7 @@ def permissions_accepted(*fsperms):
     return wrapper
 
 
-def anonymous_user_required(f):
+def anonymous_user_required(f: DecoratedView) -> DecoratedView:
     """Decorator which requires that caller NOT be logged in.
     If a logged in user accesses an endpoint protected with this decorator
     they will be redirected to the *SECURITY_POST_LOGIN_VIEW*.
@@ -597,4 +617,4 @@ def anonymous_user_required(f):
                 return redirect(get_url(_security.post_login_view))
         return f(*args, **kwargs)
 
-    return wrapper
+    return t.cast(DecoratedView, wrapper)
