@@ -17,13 +17,13 @@ import pytest
 from flask import Blueprint
 
 from flask_security import uia_email_mapper
-from flask_security.utils import hash_data
 
 from tests.test_utils import (
     authenticate,
-    json_authenticate,
+    get_auth_token_version_3x,
     get_num_queries,
     hash_password,
+    json_authenticate,
     logout,
     populate_data,
     verify_token,
@@ -153,6 +153,15 @@ def test_login_form(client):
     assert re.search(b'<input[^>]*type="email"[^>]*>', response.data)
 
 
+@pytest.mark.settings(username_enable=True)
+def test_login_form_username(client):
+    # If USERNAME_ENABLE is set then login form should have a StringField, not Email
+    response = client.post("/login", data={"email": "matt@lp.com"})
+    assert b"matt@lp.com" in response.data
+    # Should be a string field - not html5 email input type
+    assert not re.search(b'<input[^>]*type="email"[^>]*>', response.data)
+
+
 def test_unprovided_username(client, get_message):
     response = authenticate(client, "")
     assert get_message("EMAIL_NOT_PROVIDED") in response.data
@@ -711,7 +720,7 @@ def test_login_info(client):
 @pytest.mark.registerable()
 @pytest.mark.settings(post_login_view="/anon_required")
 def test_anon_required(client, get_message):
-    """ If logged in, should get 'anonymous_user_required' redirect """
+    """If logged in, should get 'anonymous_user_required' redirect"""
     response = authenticate(client, follow_redirects=False)
     response = client.get("/register")
     assert "location" in response.headers
@@ -721,7 +730,7 @@ def test_anon_required(client, get_message):
 @pytest.mark.registerable()
 @pytest.mark.settings(post_login_view="/anon_required")
 def test_anon_required_json(client, get_message):
-    """ If logged in, should get 'anonymous_user_required' response """
+    """If logged in, should get 'anonymous_user_required' response"""
     authenticate(client, follow_redirects=False)
     response = client.get("/register", headers={"Accept": "application/json"})
     assert response.status_code == 400
@@ -751,20 +760,14 @@ def test_change_uniquifier(app, client_nc):
     verify_token(client_nc, token)
 
 
-def test_verifying_token_from_version_3x(app, client_nc):
+def test_verifying_token_from_version_3x(in_app_context):
     """
     Check token generated with flask security 3.x, which has different form
     than token from version 4.0.0, can be verified
     """
 
-    def get_auth_token_version_3x(app, user):
-        """
-        Copy of algorithm that generated user token in version 3.x
-        """
-        data = [str(user.id), hash_data(user.password)]
-        if hasattr(user, "fs_uniquifier"):
-            data.append(user.fs_uniquifier)
-        return app.security.remember_token_serializer.dumps(data)
+    app = in_app_context
+    populate_data(app)
 
     with app.test_request_context("/"):
         user = app.security.datastore.find_user(email="matt@lp.com")
@@ -937,3 +940,25 @@ def test_no_get_auth_token(app, client):
         headers={"Content-Type": "application/json"},
     )
     assert "authentication_token" not in response.json["response"]["user"]
+
+
+def test_auth_token_decorator(in_app_context):
+    """
+    Test accessing endpoint decorated with auth_token_required
+    when using token generated by flask security 3.x algorithm
+    """
+
+    app = in_app_context
+    populate_data(app)
+    client_nc = app.test_client(use_cookies=False)
+
+    with app.test_request_context("/"):
+
+        user = app.security.datastore.find_user(email="matt@lp.com")
+        token = get_auth_token_version_3x(app, user)
+
+        response = client_nc.get(
+            "/token",
+            headers={"Content-Type": "application/json", "Authentication-Token": token},
+        )
+        assert response.status_code == 200
