@@ -1,4 +1,3 @@
-# -*- coding: utf-8 -*-
 """
     test_confirmable
     ~~~~~~~~~~~~~~~~
@@ -6,27 +5,29 @@
     Confirmable tests
 """
 
+import re
 import time
+from urllib.parse import parse_qsl, urlsplit
 
 import pytest
 from flask import Flask
-from utils import authenticate, logout
 
 from flask_security.core import UserMixin
 from flask_security.confirmable import generate_confirmation_token
 from flask_security.signals import confirm_instructions_sent, user_confirmed
-from flask_security.utils import capture_flashes, capture_registrations, string_types
 
-try:
-    from urlparse import parse_qsl, urlsplit
-except ImportError:  # pragma: no cover
-    from urllib.parse import parse_qsl, urlsplit
+from tests.test_utils import (
+    authenticate,
+    capture_flashes,
+    capture_registrations,
+    logout,
+)
 
 pytestmark = pytest.mark.confirmable()
 
 
 @pytest.mark.registerable()
-def test_confirmable_flag(app, client, sqlalchemy_datastore, get_message):
+def test_confirmable_flag(app, clients, get_message):
     recorded_confirms = []
     recorded_instructions_sent = []
 
@@ -37,30 +38,31 @@ def test_confirmable_flag(app, client, sqlalchemy_datastore, get_message):
         recorded_confirms.append(user)
 
     @confirm_instructions_sent.connect_via(app)
-    def on_instructions_sent(app, user, token):
+    def on_instructions_sent(app, **kwargs):
         assert isinstance(app, Flask)
-        assert isinstance(user, UserMixin)
-        assert isinstance(token, string_types)
-        recorded_instructions_sent.append(user)
+        assert isinstance(kwargs["user"], UserMixin)
+        assert isinstance(kwargs["token"], str)
+        assert isinstance(kwargs["confirmation_token"], str)
+        recorded_instructions_sent.append(kwargs["user"])
 
     # Test login before confirmation
     email = "dude@lp.com"
 
     with capture_registrations() as registrations:
         data = dict(email=email, password="awesome sunset", next="")
-        response = client.post("/register", data=data)
+        response = clients.post("/register", data=data)
 
     assert response.status_code == 302
 
-    response = authenticate(client, email=email, password="awesome sunset")
+    response = authenticate(clients, email=email, password="awesome sunset")
     assert get_message("CONFIRMATION_REQUIRED") in response.data
 
     # Test invalid token
-    response = client.get("/confirm/bogus", follow_redirects=True)
+    response = clients.get("/confirm/bogus", follow_redirects=True)
     assert get_message("INVALID_CONFIRMATION_TOKEN") in response.data
 
     # Test JSON
-    response = client.post(
+    response = clients.post(
         "/confirm",
         json=dict(email="matt@lp.com"),
         headers={"Content-Type": "application/json"},
@@ -71,22 +73,22 @@ def test_confirmable_flag(app, client, sqlalchemy_datastore, get_message):
     assert len(recorded_instructions_sent) == 1
 
     # Test ask for instructions with invalid email
-    response = client.post("/confirm", data=dict(email="bogus@bogus.com"))
+    response = clients.post("/confirm", data=dict(email="bogus@bogus.com"))
     assert get_message("USER_DOES_NOT_EXIST") in response.data
 
     # Test resend instructions
-    response = client.post("/confirm", data=dict(email=email))
+    response = clients.post("/confirm", data=dict(email=email))
     assert get_message("CONFIRMATION_REQUEST", email=email) in response.data
     assert len(recorded_instructions_sent) == 2
 
     # Test confirm
     token = registrations[0]["confirm_token"]
-    response = client.get("/confirm/" + token, follow_redirects=True)
+    response = clients.get("/confirm/" + token, follow_redirects=True)
     assert get_message("EMAIL_CONFIRMED") in response.data
     assert len(recorded_confirms) == 1
 
     # Test already confirmed
-    response = client.get("/confirm/" + token, follow_redirects=True)
+    response = clients.get("/confirm/" + token, follow_redirects=True)
     assert get_message("ALREADY_CONFIRMED") in response.data
     assert len(recorded_instructions_sent) == 2
 
@@ -95,35 +97,92 @@ def test_confirmable_flag(app, client, sqlalchemy_datastore, get_message):
     with app.app_context():
         user = registrations[0]["user"]
         expired_token = generate_confirmation_token(user)
-    response = client.get("/confirm/" + expired_token, follow_redirects=True)
+    response = clients.get("/confirm/" + expired_token, follow_redirects=True)
     assert get_message("ALREADY_CONFIRMED") in response.data
     assert len(recorded_instructions_sent) == 2
 
     # Test already confirmed when asking for confirmation instructions
-    logout(client)
+    logout(clients)
 
-    response = client.get("/confirm")
+    response = clients.get("/confirm")
     assert response.status_code == 200
 
-    response = client.post("/confirm", data=dict(email=email))
+    response = clients.post("/confirm", data=dict(email=email))
     assert get_message("ALREADY_CONFIRMED") in response.data
 
-    # Test user was deleted before confirmation
+    # Test if user was deleted before confirmation
     with capture_registrations() as registrations:
-        data = dict(email="mary@lp.com", password="awesome sunset", next="")
-        client.post("/register", data=data)
+        data = dict(email="mary27@lp.com", password="awesome sunset", next="")
+        clients.post("/register", data=data)
 
     user = registrations[0]["user"]
     token = registrations[0]["confirm_token"]
 
     with app.app_context():
-        sqlalchemy_datastore.delete(user)
-        sqlalchemy_datastore.commit()
+        app.security.datastore.delete(user)
+        app.security.datastore.commit()
+        if hasattr(app.security.datastore.db, "close_db"):
+            app.security.datastore.db.close_db(None)
 
-    response = client.get("/confirm/" + token, follow_redirects=True)
+    response = clients.get("/confirm/" + token, follow_redirects=True)
     assert get_message("INVALID_CONFIRMATION_TOKEN") in response.data
 
 
+@pytest.mark.registerable()
+def test_confirmation_template(app, client, get_message):
+    # Check contents of email template - this uses a test template
+    # in order to check all context vars since the default template
+    # doesn't have all of them.
+
+    recorded_tokens_sent = []
+
+    @confirm_instructions_sent.connect_via(app)
+    def on_instructions_sent(app, **kwargs):
+        recorded_tokens_sent.append(kwargs["confirmation_token"])
+
+    with capture_registrations() as registrations:
+        with app.mail.record_messages() as outbox:
+            data = dict(email="mary@lp.com", password="awesome sunset", next="")
+            # Register - this will use the welcome template
+            client.post("/register", data=data, follow_redirects=True)
+            # Explicitly ask for confirmation -
+            # this will use the confirmation_instructions template
+            client.post("/confirm", data=dict(email="mary@lp.com"))
+        assert len(outbox) == 2
+        # check registration email
+        matcher = re.findall(r"\w+:.*", outbox[0].body, re.IGNORECASE)
+        # should be 4 - link, email, token, config item
+        assert matcher[1].split(":")[1] == "mary@lp.com"
+        assert matcher[2].split(":")[1] == registrations[0]["confirm_token"]
+        assert matcher[2].split(":")[1] == registrations[0]["confirmation_token"]
+        assert matcher[3].split(":")[1] == app.config["SECURITY_CONFIRM_URL"]
+
+        # check confirmation email
+        matcher = re.findall(r"\w+:.*", outbox[1].body, re.IGNORECASE)
+        # should be 4 - link, email, token, config item
+        assert matcher[1].split(":")[1] == "mary@lp.com"
+        assert matcher[2].split(":")[1] == recorded_tokens_sent[0]
+        token = matcher[2].split(":")[1]
+        assert token == recorded_tokens_sent[0]
+        assert matcher[3].split(":")[1] == app.config["SECURITY_CONFIRM_URL"]
+
+        # check link
+        _, link = matcher[0].split(":", 1)
+        response = client.get(link, follow_redirects=True)
+        assert get_message("EMAIL_CONFIRMED") in response.data
+
+
+@pytest.mark.registerable()
+@pytest.mark.settings(requires_confirmation_error_view="/confirm")
+def test_requires_confirmation_error_redirect(app, clients):
+    data = dict(email="jyl@lp.com", password="awesome sunset")
+    response = clients.post("/register", data=data)
+
+    response = authenticate(clients, **data, follow_redirects=True)
+    assert b"send_confirmation_form" in response.data
+    assert b"jyl@lp.com" in response.data
+
+
 @pytest.mark.registerable()
 @pytest.mark.settings(confirm_email_within="1 milliseconds")
 def test_expired_confirmation_token(client, get_message):
@@ -173,7 +232,7 @@ def test_login_when_unconfirmed(client, get_message):
 
 @pytest.mark.registerable()
 def test_no_auth_token(client_nc):
-    """ Make sure that register doesn't return Authentication Token
+    """Make sure that register doesn't return Authentication Token
     if user isn't confirmed.
     """
     response = client_nc.post(
@@ -183,13 +242,13 @@ def test_no_auth_token(client_nc):
     )
     assert response.status_code == 200
     user = response.json["response"]["user"]
-    assert len(user) == 2 and all(k in user for k in ["id", "last_update"])
+    assert len(user) == 2 and all(k in user for k in ["email", "last_update"])
 
 
 @pytest.mark.registerable()
 @pytest.mark.settings(login_without_confirmation=True)
 def test_auth_token_unconfirmed(client_nc):
-    """ Make sure that register returns Authentication Token
+    """Make sure that register returns Authentication Token
     if user isn't confirmed, but the 'login_without_confirmation' flag is set.
     """
     response = client_nc.post(
@@ -200,14 +259,14 @@ def test_auth_token_unconfirmed(client_nc):
     assert response.status_code == 200
     user = response.json["response"]["user"]
     assert len(user) == 3 and all(
-        k in user for k in ["id", "last_update", "authentication_token"]
+        k in user for k in ["email", "last_update", "authentication_token"]
     )
 
 
 @pytest.mark.registerable()
 @pytest.mark.settings(login_without_confirmation=True, auto_login_after_confirm=False)
 def test_confirmation_different_user_when_logged_in_no_auto(client, get_message):
-    """ Default - AUTO_LOGIN == false so shouldn't log in second user. """
+    """Default - AUTO_LOGIN == false so shouldn't log in second user."""
     e1 = "dude@lp.com"
     e2 = "lady@lp.com"
 
@@ -340,7 +399,7 @@ def test_spa_get(app, client):
     confirm_error_view="/confirm-error",
 )
 def test_spa_get_bad_token(app, client, get_message):
-    """ Test expired and invalid token"""
+    """Test expired and invalid token"""
     with capture_flashes() as flashes:
         with capture_registrations() as registrations:
             response = client.post(
@@ -358,8 +417,8 @@ def test_spa_get_bad_token(app, client, get_message):
         assert "localhost:8081" == split.netloc
         assert "/confirm-error" == split.path
         qparams = dict(parse_qsl(split.query))
-        assert len(qparams) == 2
-        assert all(k in qparams for k in ["email", "error"])
+        assert all(k in qparams for k in ["email", "error", "identity"])
+        assert qparams["identity"] == "dude@lp.com"
 
         msg = get_message(
             "CONFIRMATION_EXPIRED", within="1 milliseconds", email="dude@lp.com"
@@ -390,7 +449,7 @@ def test_spa_get_bad_token(app, client, get_message):
 @pytest.mark.registerable()
 @pytest.mark.settings(two_factor_required=True)
 def test_two_factor(app, client):
-    """ If two-factor is enabled, the confirm shouldn't login, but start the
+    """If two-factor is enabled, the confirm shouldn't login, but start the
     2-factor setup.
     """
     with capture_registrations() as registrations:
@@ -432,3 +491,54 @@ def test_two_factor_json(app, client, get_message):
     assert response.status_code == 200
     assert response.json["response"]["tf_required"]
     assert response.json["response"]["tf_state"] == "setup_from_login"
+
+
+@pytest.mark.registerable()
+@pytest.mark.settings(
+    user_identity_attributes=[{"username": {"mapper": lambda x: x}}],
+)
+def test_email_not_identity(app, sqlalchemy_datastore, get_message):
+    # Test that can register/confirm with email even if it isn't an IDENTITY_ATTRIBUTE
+    from flask_security import ConfirmRegisterForm, Security, unique_identity_attribute
+    from wtforms import StringField, validators
+
+    class MyRegisterForm(ConfirmRegisterForm):
+        username = StringField(
+            "Username",
+            validators=[validators.data_required(), unique_identity_attribute],
+        )
+
+    app.config["SECURITY_CONFIRM_REGISTER_FORM"] = MyRegisterForm
+    security = Security(datastore=sqlalchemy_datastore)
+    security.init_app(app)
+
+    client = app.test_client()
+
+    with capture_registrations() as registrations:
+        data = dict(email="mary2@lp.com", username="mary", password="awesome sunset")
+        response = client.post("/register", data=data, follow_redirects=True)
+        assert b"mary2@lp.com" in response.data
+
+    token = registrations[0]["confirm_token"]
+    response = client.get("/confirm/" + token, headers={"Accept": "application/json"})
+    assert response.status_code == 302
+    assert response.location == "http://localhost/"
+
+    logout(client)
+
+    # check that username must be unique
+    data = dict(email="mary4@lp.com", username="mary", password="awesome sunset")
+    response = client.post(
+        "/register", data=data, headers={"Accept": "application/json"}
+    )
+    assert response.status_code == 400
+    assert "is already associated" in response.json["response"]["errors"]["username"][0]
+
+    # log in with username - this uses the age-old hack that although the form's
+    # input label says "email" - it in fact will accept any identity attribute.
+    response = client.post(
+        "/login",
+        data=dict(email="mary", password="awesome sunset"),
+        follow_redirects=True,
+    )
+    assert b"<p>Welcome mary</p>" in response.data
