@@ -5,7 +5,7 @@
     Lots of tests
 
     :copyright: (c) 2012 by Matt Wright.
-    :copyright: (c) 2019-2020 by J. Christopher Wagner (jwag).
+    :copyright: (c) 2019-2021 by J. Christopher Wagner (jwag).
     :license: MIT, see LICENSE for more details.
 """
 
@@ -17,9 +17,12 @@ import os.path
 import pkg_resources
 import sys
 import time
+import typing as t
 
 import pytest
 
+from wtforms.validators import DataRequired, Length
+
 from tests.test_utils import (
     authenticate,
     capture_flashes,
@@ -31,7 +34,7 @@ from tests.test_utils import (
     populate_data,
 )
 
-from flask import abort, request, Response
+from flask import Flask, abort, request, Response
 from flask_security import Security
 from flask_security.forms import (
     ChangePasswordForm,
@@ -58,9 +61,13 @@ from flask_security.utils import (
     hash_data,
     send_mail,
     uia_phone_mapper,
+    validate_redirect_url,
     verify_hash,
 )
 
+if t.TYPE_CHECKING:  # pragma: no cover
+    from flask.testing import FlaskClient
+
 
 @pytest.mark.recoverable()
 def test_my_mail_util(app, sqlalchemy_datastore):
@@ -391,10 +398,12 @@ def test_form_required(app, sqlalchemy_datastore):
 
 
 def test_form_required_local_message(app, sqlalchemy_datastore):
-    """ Test having a local message (not xlatable and not part of MSG_ config."""
+    """Test having a local message (not xlatable and not part of MSG_ config."""
+
+    msg = "hi! did you forget me?"
 
     class MyLoginForm(LoginForm):
-        myfield = StringField("My Custom Field", validators=[Required(message="hi")])
+        myfield = StringField("My Custom Field", validators=[Required(message=msg)])
 
     app.config["SECURITY_LOGIN_FORM"] = MyLoginForm
 
@@ -406,16 +415,20 @@ def test_form_required_local_message(app, sqlalchemy_datastore):
     response = client.post("/login", content_type="application/json")
     assert response.status_code == 400
     assert b"myfield" in response.data
+    assert msg.encode("utf-8") in response.data
+    # WTforms 2.x incorrectly catches ValueError and sets that as the form error.
+    # Our config_value routine raises ValueError for missing config items..
+    assert b"Key" not in response.data
 
 
-@pytest.mark.babel(False)
-def test_without_babel(client):
-    # This isn't really 'without' babel - it is without initializing babel
-    with pytest.raises(ValueError):
-        client.get("/login")
+def test_without_babel(app, client):
+    # Test if babel modules exist but we don't init babel - things still work
+    app.config["BABEL_DEFAULT_LOCALE"] = "fr_FR"
+    response = client.get("/login")
+    assert response.status_code == 200
 
 
-def test_no_email_sender(app):
+def test_no_email_sender(app, sqlalchemy_datastore):
     """Verify that if SECURITY_EMAIL_SENDER is default
     (which is a local proxy) that send_mail picks up MAIL_DEFAULT_SENDER.
     """
@@ -426,17 +439,40 @@ def test_no_email_sender(app):
             self.email = email
 
     security = Security()
-    security.init_app(app)
+    security.init_app(app, sqlalchemy_datastore)
 
     with app.app_context():
         app.try_trigger_before_first_request_functions()
         user = TestUser("matt@lp.com")
         with app.mail.record_messages() as outbox:
             send_mail("Test Default Sender", user.email, "welcome", user=user)
-        assert 1 == len(outbox)
-        assert "test@testme.com" == outbox[0].sender
+            assert 1 == len(outbox)
+            assert "test@testme.com" == outbox[0].sender
 
 
+def test_sender_tuple(app, sqlalchemy_datastore):
+    """Verify that if sender is a (name, address) tuple,
+    in the received email sender is properly formatted as "name <address>"
+    """
+    app.config["MAIL_DEFAULT_SENDER"] = ("Test User", "test@testme.com")
+
+    class TestUser:
+        def __init__(self, email):
+            self.email = email
+
+    security = Security()
+    security.init_app(app, sqlalchemy_datastore)
+
+    with app.app_context():
+        app.try_trigger_before_first_request_functions()
+        user = TestUser("matt@lp.com")
+        with app.mail.record_messages() as outbox:
+            send_mail("Test Tuple Sender", user.email, "welcome", user=user)
+            assert 1 == len(outbox)
+            assert "Test User <test@testme.com>" == outbox[0].sender
+
+
+@pytest.mark.babel()
 def test_xlation(app, client):
     app.config["BABEL_DEFAULT_LOCALE"] = "fr_FR"
     assert check_xlation(app, "fr_FR"), "You must run python setup.py compile_catalog"
@@ -449,8 +485,16 @@ def test_xlation(app, client):
     assert b"Bienvenue matt@lp.com" in response.data
 
 
+@pytest.mark.babel()
 def test_myxlation(app, sqlalchemy_datastore, pytestconfig):
     # Test changing a single MSG and having an additional translation dir
+    # Flask-BabelEx doesn't support lists of directories..
+    try:
+        import flask_babelex  # noqa: F401
+
+        pytest.skip("Flask-BabelEx doesn't support lists of translations")
+    except ImportError:
+        pass
 
     i18n_dirname = [
         pkg_resources.resource_filename("flask_security", "translations"),
@@ -470,10 +514,11 @@ def test_myxlation(app, sqlalchemy_datastore, pytestconfig):
     assert b"Passe - no-worky" in response.data
 
 
-def test_form_labels(app):
+@pytest.mark.babel()
+def test_form_labels(app, sqlalchemy_datastore):
     app.config["BABEL_DEFAULT_LOCALE"] = "fr_FR"
     app.security = Security()
-    app.security.init_app(app)
+    app.security.init_app(app, sqlalchemy_datastore)
     assert check_xlation(app, "fr_FR"), "You must run python setup.py compile_catalog"
 
     with app.test_request_context():
@@ -496,7 +541,39 @@ def test_form_labels(app):
         assert str(form.submit.label.text) == "Changer le mot de passe"
 
 
+@pytest.mark.babel()
+def test_wtform_xlation(app, sqlalchemy_datastore):
+    # Make sure wtform xlations work
+    class MyLoginForm(LoginForm):
+        fixed_length = StringField(
+            "FixedLength", validators=[DataRequired(), Length(3, 3)]
+        )
+
+    app.config["BABEL_DEFAULT_LOCALE"] = "fr_FR"
+    app.security = Security()
+    app.security.init_app(app, datastore=sqlalchemy_datastore, login_form=MyLoginForm)
+    assert check_xlation(app, "fr_FR"), "You must run python setup.py compile_catalog"
+
+    client = app.test_client()
+    response = client.get("/login")
+    assert b'<label for="password">Mot de passe</label>' in response.data
+    data = dict(
+        email="matt@lp.com", password="", remember="y", fixed_length="waytoolong"
+    )
+    response = client.post(
+        "/login", json=data, headers={"Content-Type": "application/json"}
+    )
+    assert response.status_code == 400
+    flerror = response.json["response"]["errors"]["fixed_length"][0]
+    # This is completely dependent on WTforms translations....
+    assert (
+        flerror == "Le doit contenir exactement 3 caractères."
+        or flerror == "Le champ doit contenir exactement 3 caractères."
+    )
+
+
 @pytest.mark.changeable()
+@pytest.mark.babel()
 def test_per_request_xlate(app, client):
     from flask import request, session
 
@@ -572,7 +649,7 @@ def test_zxcvbn_xlate(app):
 
 @pytest.mark.skipif(sys.version_info < (3, 0), reason="requires python3 or higher")
 @pytest.mark.settings(password_check_breached="strict")
-def test_breached(app):
+def test_breached(app, sqlalchemy_datastore):
 
     # partial response from: https://api.pwnedpasswords.com/range/07003
     pwned_response = b"AF5A73CD3CBCFDCD12B0B68CB7930F3E888:2\r\n\
@@ -584,7 +661,7 @@ B25C03CFBE4CBF19E0F4889711C9A488E5D:2\r\n\
 B3902FD808DCA504AAAD30F3C14BD3ACE7C:10"
 
     app.security = Security()
-    app.security.init_app(app)
+    app.security.init_app(app, sqlalchemy_datastore)
     with app.test_request_context():
         with mock.patch("urllib.request.urlopen") as mock_urlopen:
             mock_urlopen.return_value.__enter__.return_value.read.return_value = (
@@ -601,7 +678,7 @@ B3902FD808DCA504AAAD30F3C14BD3ACE7C:10"
     password_breached_count=16,
     password_complexity_checker="zxcvbn",
 )
-def test_breached_cnt(app):
+def test_breached_cnt(app, sqlalchemy_datastore):
 
     # partial response from: https://api.pwnedpasswords.com/range/07003
     pwned_response = b"AF5A73CD3CBCFDCD12B0B68CB7930F3E888:2\r\n\
@@ -613,7 +690,7 @@ B25C03CFBE4CBF19E0F4889711C9A488E5D:2\r\n\
 B3902FD808DCA504AAAD30F3C14BD3ACE7C:10"
 
     app.security = Security()
-    app.security.init_app(app)
+    app.security.init_app(app, sqlalchemy_datastore)
     with app.test_request_context():
         with mock.patch("urllib.request.urlopen") as mock_urlopen:
             mock_urlopen.return_value.__enter__.return_value.read.return_value = (
@@ -627,11 +704,11 @@ B3902FD808DCA504AAAD30F3C14BD3ACE7C:10"
 
 @pytest.mark.skip
 @pytest.mark.settings(password_check_breached="strict")
-def test_breached_real(app):
-    """ Actually go out to internet.. """
+def test_breached_real(app, sqlalchemy_datastore):
+    """Actually go out to internet.."""
 
     app.security = Security()
-    app.security.init_app(app)
+    app.security.init_app(app, sqlalchemy_datastore)
     with app.test_request_context():
         pbad, pnorm = app.security._password_util.validate("flaskflask", True)
         assert len(pbad) == 1
@@ -639,7 +716,7 @@ def test_breached_real(app):
 
 
 def test_json_error_response_string():
-    """ Unit test for correct response when a string is given. """
+    """Unit test for correct response when a string is given."""
     error_msg = "This is an error!"
     response = json_error_response(errors=error_msg)
     assert "error" in response
@@ -648,7 +725,7 @@ def test_json_error_response_string():
 
 
 def test_json_error_response_dict():
-    """ Unit test for correct response when a dict is given. """
+    """Unit test for correct response when a dict is given."""
     error_msg = {
         "e-mail": "The e-mail address is already in the system.",
         "name": "The name is too long.",
@@ -660,7 +737,7 @@ def test_json_error_response_dict():
 
 
 def test_json_error_response_typeerror():
-    """ Unit test for checking for error raising. """
+    """Unit test for checking for error raising."""
     error_msg = ("tuple",)
     with pytest.raises(TypeError):
         json_error_response(errors=error_msg)
@@ -692,7 +769,7 @@ def test_method_view(app, client):
     assert b"Hi view" in response.data
 
 
-def test_phone_util_override(app):
+def test_phone_util_override(app, sqlalchemy_datastore):
     class MyPhoneUtil:
         def __init__(self, app):
             pass
@@ -704,14 +781,16 @@ def test_phone_util_override(app):
             return "very-canonical"
 
     app.security = Security()
-    app.security.init_app(app, phone_util_cls=MyPhoneUtil)
+    app.security.init_app(app, sqlalchemy_datastore, phone_util_cls=MyPhoneUtil)
 
     with app.app_context():
         assert uia_phone_mapper("55") == "very-canonical"
 
 
-def test_authn_freshness(app, client, get_message):
-    """ Test freshness using default reauthn_handler """
+def test_authn_freshness(
+    app: "Flask", client: "FlaskClient", get_message: t.Callable[..., bytes]
+) -> None:
+    """Test freshness using default reauthn_handler"""
 
     @auth_required(within=30, grace=0)
     def myview():
@@ -746,13 +825,13 @@ def test_authn_freshness(app, client, get_message):
     # Test json error response
     response = client.get("/myspecialview", headers={"accept": "application/json"})
     assert response.status_code == 401
-    assert response.json["response"]["error"].encode("utf-8") == get_message(
-        "REAUTHENTICATION_REQUIRED"
-    )
+    assert response.json and response.json["response"]["error"].encode(
+        "utf-8"
+    ) == get_message("REAUTHENTICATION_REQUIRED")
 
 
 def test_authn_freshness_handler(app, client, get_message):
-    """ Test with our own handler """
+    """Test with our own handler"""
 
     @app.security.reauthn_handler
     def my_reauthn(within, grace, headers=None):
@@ -960,7 +1039,7 @@ def test_verify_next(app, client, get_message):
 
 
 def test_direct_decorator(app, client, get_message):
-    """ Test/show calling the auth_required decorator directly """
+    """Test/show calling the auth_required decorator directly"""
     headers = {"Accept": "application/json", "Content-Type": "application/json"}
 
     def myview():
@@ -983,7 +1062,7 @@ def test_direct_decorator(app, client, get_message):
 
 
 def test_authn_via(app, client, get_message):
-    """ Test that we get correct fs_authn_via set in request """
+    """Test that we get correct fs_authn_via set in request"""
 
     @auth_required(within=30, grace=0)
     def myview():
@@ -1034,3 +1113,57 @@ def test_post_security_with_application_root_and_views(app, sqlalchemy_datastore
     response = client.get("/logout")
     assert response.status_code == 302
     assert response.headers["Location"] == "http://localhost/post_logout"
+
+
+@pytest.mark.settings(redirect_validate_mode="regex")
+def test_validate_redirect(app, sqlalchemy_datastore):
+    """
+    Test various possible URLs that urlsplit() shows as relative but
+    many browsers will interpret as absolute - and thus have a
+    open-redirect vulnerability. Note this vulnerability only
+    is viable if the application sets autocorrect_location_header = False
+    """
+    init_app_with_options(app, sqlalchemy_datastore)
+    with app.test_request_context("http://localhost:5001/login"):
+        assert not validate_redirect_url("\\\\\\github.com")
+        assert not validate_redirect_url(" //github.com")
+        assert not validate_redirect_url("\t//github.com")
+        assert not validate_redirect_url("//github.com")  # this is normal urlsplit
+
+
+def test_kwargs():
+    import warnings
+
+    warnings.simplefilter("error")
+    with pytest.raises(DeprecationWarning):
+        Security(myownkwarg="hi")
+
+
+def test_nodatastore(app):
+    with pytest.raises(ValueError):
+        s = Security(app)
+        s.init_app(app)
+
+
+def test_reuse_security_object(sqlalchemy_datastore):
+    # See: https://github.com/Flask-Middleware/flask-security/issues/518
+    # Let folks re-use the Security object (mostly for testing).
+    security = Security(datastore=sqlalchemy_datastore)
+
+    app = Flask(__name__)
+    app.response_class = Response
+    app.debug = True
+    app.config["SECRET_KEY"] = "secret"
+    app.config["TESTING"] = True
+
+    security.init_app(app)
+    assert hasattr(app, "login_manager")
+
+    app = Flask(__name__)
+    app.response_class = Response
+    app.debug = True
+    app.config["SECRET_KEY"] = "secret"
+    app.config["TESTING"] = True
+
+    security.init_app(app)
+    assert hasattr(app, "login_manager")
