@@ -1,4 +1,3 @@
-# -*- coding: utf-8 -*-
 """
     test_recoverable
     ~~~~~~~~~~~~~~~~
@@ -6,30 +5,27 @@
     Recoverable functionality tests
 """
 
+import re
 import time
+from urllib.parse import parse_qsl, urlsplit
 
 import pytest
 from flask import Flask
-from utils import authenticate, json_authenticate, json_logout, logout, verify_token
+from tests.test_utils import (
+    authenticate,
+    capture_flashes,
+    capture_reset_password_requests,
+    logout,
+)
 
 from flask_security.core import UserMixin
 from flask_security.forms import LoginForm
 from flask_security.signals import password_reset, reset_password_instructions_sent
-from flask_security.utils import (
-    capture_flashes,
-    capture_reset_password_requests,
-    string_types,
-)
-
-try:
-    from urlparse import parse_qsl, urlsplit
-except ImportError:  # pragma: no cover
-    from urllib.parse import parse_qsl, urlsplit
 
 pytestmark = pytest.mark.recoverable()
 
 
-def test_recoverable_flag(app, client, get_message):
+def test_recoverable_flag(app, clients, get_message):
     recorded_resets = []
     recorded_instructions_sent = []
 
@@ -38,20 +34,21 @@ def test_recoverable_flag(app, client, get_message):
         recorded_resets.append(user)
 
     @reset_password_instructions_sent.connect_via(app)
-    def on_instructions_sent(app, user, token):
+    def on_instructions_sent(app, **kwargs):
         assert isinstance(app, Flask)
-        assert isinstance(user, UserMixin)
-        assert isinstance(token, string_types)
-        recorded_instructions_sent.append(user)
+        assert isinstance(kwargs["user"], UserMixin)
+        assert isinstance(kwargs["token"], str)
+        recorded_instructions_sent.append(kwargs["user"])
 
     # Test the reset view
-    response = client.get("/reset")
+    response = clients.get("/reset")
     assert b"<h1>Send password reset instructions</h1>" in response.data
+    assert re.search(b'<input[^>]*type="email"[^>]*>', response.data)
 
     # Test submitting email to reset password creates a token and sends email
     with capture_reset_password_requests() as requests:
         with app.mail.record_messages() as outbox:
-            response = client.post(
+            response = clients.post(
                 "/reset", data=dict(email="joe@lp.com"), follow_redirects=True
             )
 
@@ -62,18 +59,18 @@ def test_recoverable_flag(app, client, get_message):
     token = requests[0]["token"]
 
     # Test view for reset token
-    response = client.get("/reset/" + token)
+    response = clients.get("/reset/" + token)
     assert b"<h1>Reset password</h1>" in response.data
 
     # Test submitting a new password but leave out confirm
-    response = client.post(
+    response = clients.post(
         "/reset/" + token, data={"password": "newpassword"}, follow_redirects=True
     )
     assert get_message("PASSWORD_NOT_PROVIDED") in response.data
     assert len(recorded_resets) == 0
 
     # Test submitting a new password
-    response = client.post(
+    response = clients.post(
         "/reset/" + token,
         data={"password": "awesome sunset", "password_confirm": "awesome sunset"},
         follow_redirects=True,
@@ -82,26 +79,26 @@ def test_recoverable_flag(app, client, get_message):
     assert get_message("PASSWORD_RESET") in response.data
     assert len(recorded_resets) == 1
 
-    logout(client)
+    logout(clients)
 
     # Test logging in with the new password
     response = authenticate(
-        client, "joe@lp.com", "awesome sunset", follow_redirects=True
+        clients, "joe@lp.com", "awesome sunset", follow_redirects=True
     )
     assert b"Welcome joe@lp.com" in response.data
 
-    logout(client)
+    logout(clients)
 
     # Test invalid email
-    response = client.post(
+    response = clients.post(
         "/reset", data=dict(email="bogus@lp.com"), follow_redirects=True
     )
     assert get_message("USER_DOES_NOT_EXIST") in response.data
 
-    logout(client)
+    logout(clients)
 
     # Test invalid token
-    response = client.post(
+    response = clients.post(
         "/reset/bogus",
         data={"password": "awesome sunset", "password_confirm": "awesome sunset"},
         follow_redirects=True,
@@ -114,7 +111,7 @@ def test_recoverable_flag(app, client, get_message):
         "BZEw_Q.lQyo3npdPZtcJ_sNHVHP103syjM"
         "&url_id=fbb89a8328e58c181ea7d064c2987874bc54a23d"
     )
-    response = client.post(
+    response = clients.post(
         "/reset/" + token,
         data={"password": "newpassword", "password_confirm": "newpassword"},
         follow_redirects=True,
@@ -122,6 +119,20 @@ def test_recoverable_flag(app, client, get_message):
     assert get_message("INVALID_RESET_PASSWORD_TOKEN") in response.data
 
 
+@pytest.mark.confirmable()
+@pytest.mark.registerable()
+@pytest.mark.settings(requires_confirmation_error_view="/confirm")
+def test_requires_confirmation_error_redirect(app, clients):
+    data = dict(email="jyl@lp.com", password="awesome sunset")
+    clients.post("/register", data=data)
+
+    response = clients.post(
+        "/reset", data=dict(email="jyl@lp.com"), follow_redirects=True
+    )
+    assert b"send_confirmation_form" in response.data
+    assert b"jyl@lp.com" in response.data
+
+
 @pytest.mark.settings()
 def test_recoverable_json(app, client, get_message):
     recorded_resets = []
@@ -132,8 +143,8 @@ def test_recoverable_json(app, client, get_message):
         recorded_resets.append(user)
 
     @reset_password_instructions_sent.connect_via(app)
-    def on_instructions_sent(app, user, token):
-        recorded_instructions_sent.append(user)
+    def on_instructions_sent(app, **kwargs):
+        recorded_instructions_sent.append(kwargs["user"])
 
     with capture_flashes() as flashes:
         # Test reset password creates a token and sends email
@@ -181,7 +192,7 @@ def test_recoverable_json(app, client, get_message):
         )
         assert all(
             k in response.json["response"]["user"]
-            for k in ["id", "authentication_token"]
+            for k in ["email", "authentication_token"]
         )
         assert len(recorded_resets) == 1
 
@@ -196,7 +207,7 @@ def test_recoverable_json(app, client, get_message):
         )
         assert all(
             k in response.json["response"]["user"]
-            for k in ["id", "authentication_token"]
+            for k in ["email", "authentication_token"]
         )
 
         logout(client)
@@ -222,6 +233,65 @@ def test_recoverable_json(app, client, get_message):
     assert len(flashes) == 0
 
 
+def test_recoverable_template(app, client, get_message):
+    # Check contents of email template - this uses a test template
+    # in order to check all context vars since the default template
+    # doesn't have all of them.
+    with capture_reset_password_requests() as resets:
+        with app.mail.record_messages() as outbox:
+            response = client.post(
+                "/reset", data=dict(email="joe@lp.com"), follow_redirects=True
+            )
+        assert len(outbox) == 1
+        matcher = re.findall(r"\w+:.*", outbox[0].body, re.IGNORECASE)
+        # should be 4 - link, email, token, config item
+        assert matcher[1].split(":")[1] == "joe@lp.com"
+        assert matcher[2].split(":")[1] == resets[0]["reset_token"]
+        assert matcher[3].split(":")[1] == app.config["SECURITY_CONFIRM_URL"]
+
+        # check link
+        link = matcher[0].split(":", 1)[1]
+        response = client.get(link, follow_redirects=True)
+        assert b"Reset Password" in response.data
+
+
+def test_recover_invalidates_session(app, client):
+    # Make sure that if we reset our password - prior sessions are invalidated.
+
+    other_client = app.test_client()
+    authenticate(other_client)
+    response = other_client.get("/profile", follow_redirects=True)
+    assert b"Profile Page" in response.data
+
+    # use normal client to reset password
+    with capture_reset_password_requests() as requests:
+        response = client.post(
+            "/reset",
+            json=dict(email="matt@lp.com"),
+            headers={"Content-Type": "application/json"},
+        )
+        assert response.headers["Content-Type"] == "application/json"
+
+    assert response.status_code == 200
+    token = requests[0]["token"]
+
+    # Test submitting a new password
+    response = client.post(
+        "/reset/" + token + "?include_auth_token",
+        json=dict(password="awesome sunset", password_confirm="awesome sunset"),
+        headers={"Content-Type": "application/json"},
+    )
+    assert all(
+        k in response.json["response"]["user"]
+        for k in ["email", "authentication_token"]
+    )
+
+    # try to access protected endpoint with old session - shouldn't work
+    response = other_client.get("/profile")
+    assert response.status_code == 302
+    assert response.headers["Location"] == "http://localhost/login?next=%2Fprofile"
+
+
 def test_login_form_description(sqlalchemy_app):
     app = sqlalchemy_app()
     with app.test_request_context("/login"):
@@ -413,7 +483,7 @@ def test_spa_get(app, client):
     reset_error_view="/reset-error",
 )
 def test_spa_get_bad_token(app, client, get_message):
-    """ Test expired and invalid token"""
+    """Test expired and invalid token"""
     with capture_flashes() as flashes:
         with capture_reset_password_requests() as requests:
             response = client.post(
@@ -432,8 +502,7 @@ def test_spa_get_bad_token(app, client, get_message):
         assert "localhost:8081" == split.netloc
         assert "/reset-error" == split.path
         qparams = dict(parse_qsl(split.query))
-        assert len(qparams) == 2
-        assert all(k in qparams for k in ["email", "error"])
+        assert all(k in qparams for k in ["email", "error", "identity"])
 
         msg = get_message(
             "PASSWORD_RESET_EXPIRED", within="1 milliseconds", email="joe@lp.com"
@@ -460,41 +529,6 @@ def test_spa_get_bad_token(app, client, get_message):
     assert len(flashes) == 0
 
 
-@pytest.mark.settings(backwards_compat_auth_token_invalid=True)
-def test_bc_password(app, client_nc):
-    # Test behavior of BACKWARDS_COMPAT_AUTH_TOKEN_INVALID
-    response = json_authenticate(client_nc, email="joe@lp.com")
-    token = response.json["response"]["user"]["authentication_token"]
-    verify_token(client_nc, token)
-    json_logout(client_nc, token)
-
-    with capture_reset_password_requests() as requests:
-        response = client_nc.post(
-            "/reset",
-            json=dict(email="joe@lp.com"),
-            headers={"Content-Type": "application/json"},
-        )
-        assert response.status_code == 200
-
-    reset_token = requests[0]["token"]
-
-    data = dict(password="awesome sunset", password_confirm="awesome sunset")
-    response = client_nc.post(
-        "/reset/" + reset_token + "?include_auth_token=1",
-        json=data,
-        headers={"Content-Type": "application/json"},
-    )
-    assert response.status_code == 200
-    assert "authentication_token" in response.json["response"]["user"]
-
-    # changing password should have rendered existing auth tokens invalid
-    verify_token(client_nc, token, status=401)
-
-    # but new auth token should work
-    token = response.json["response"]["user"]["authentication_token"]
-    verify_token(client_nc, token)
-
-
 @pytest.mark.settings(password_complexity_checker="zxcvbn")
 def test_easy_password(client, get_message):
     with capture_reset_password_requests() as requests:
@@ -524,3 +558,11 @@ def test_reset_inactive(client, get_message):
         headers={"Content-Type": "application/json"},
     )
     assert response.status_code == 400
+
+
+def test_email_normalization(client, get_message):
+    response = client.post(
+        "/reset", data=dict(email="joe@LP.COM"), follow_redirects=True
+    )
+    assert response.status_code == 200
+    assert get_message("PASSWORD_RESET_REQUEST", email="joe@lp.com") in response.data
