@@ -1,50 +1,56 @@
-# -*- coding: utf-8 -*-
 """
     test_unified_signin
     ~~~~~~~~~~~~~~~~~~~
 
     Unified signin tests
 
-    :copyright: (c) 2019-2020 by J. Christopher Wagner (jwag).
+    :copyright: (c) 2019-2021 by J. Christopher Wagner (jwag).
     :license: MIT, see LICENSE for more details.
 
 """
 
+import base64
 from contextlib import contextmanager
 from datetime import timedelta
-
-try:
-    from unittest.mock import Mock
-except ImportError:
-    from mock import Mock
+from passlib.totp import TOTP
 import re
 import time
+from urllib.parse import parse_qsl, urlsplit
 
 import pytest
 from flask import Flask
-from utils import SmsBadSender, SmsTestSender, authenticate, logout
+from tests.test_utils import (
+    SmsBadSender,
+    SmsTestSender,
+    authenticate,
+    capture_flashes,
+    capture_reset_password_requests,
+    logout,
+)
 
 from flask_security import (
     SmsSenderFactory,
     SQLAlchemyUserDatastore,
     UserMixin,
     uia_email_mapper,
+    uia_phone_mapper,
     us_profile_changed,
     us_security_token_sent,
     user_authenticated,
 )
-from flask_security.utils import capture_flashes, capture_reset_password_requests
 
-try:
-    from urlparse import parse_qsl, urlsplit
-except ImportError:  # pragma: no cover
-    from urllib.parse import parse_qsl, urlsplit
+from flask_security.utils import get_identity_attributes
 
 pytestmark = pytest.mark.unified_signin()
 
 SmsSenderFactory.senders["test"] = SmsTestSender
 SmsSenderFactory.senders["bad"] = SmsBadSender
 
+UIA_EMAIL_PHONE = [
+    {"email": {"mapper": uia_email_mapper, "case_insensitive": True}},
+    {"us_phone_number": {"mapper": uia_phone_mapper}},
+]
+
 
 @contextmanager
 def capture_send_code_requests():
@@ -82,6 +88,39 @@ def us_authenticate(client, identity="matt@lp.com"):
     return response.json["response"]["user"]["authentication_token"]
 
 
+def us_tf_authenticate(app, client, json=False, validate=True, remember=False):
+    """Login/Authenticate using two factor and unified signin
+    This is the equivalent of utils:authenticate
+    """
+    prev_sms = app.config["SECURITY_SMS_SERVICE"]
+    app.config["SECURITY_SMS_SERVICE"] = "test"
+    sms_sender = SmsSenderFactory.createSender("test")
+    json_data = dict(identity="gal@lp.com", passcode="password", remember=remember)
+    response = client.post(
+        "/us-signin", json=json_data, headers={"Content-Type": "application/json"}
+    )
+
+    assert b'"code": 200' in response.data
+    app.config["SECURITY_SMS_SERVICE"] = prev_sms
+
+    if validate:
+        code = sms_sender.messages[0].split()[-1]
+        if json:
+            response = client.post(
+                "/tf-validate",
+                json=dict(code=code),
+                headers={"Content-Type": "application/json"},
+            )
+            assert b'"code": 200' in response.data
+            return response.json["response"].get("tf_validity_token", None)
+        else:
+            response = client.post(
+                "/tf-validate", data=dict(code=code), follow_redirects=True
+            )
+
+            assert response.status_code == 200
+
+
 def set_phone(app, email="matt@lp.com", phone="650-273-3780"):
     # A quick way to 'setup' SMS
     with app.test_request_context("/"):
@@ -91,7 +130,7 @@ def set_phone(app, email="matt@lp.com", phone="650-273-3780"):
         app.security.datastore.commit()
 
 
-def test_simple_signin(app, client, get_message):
+def test_simple_signin(app, clients, get_message):
     auths = []
 
     @user_authenticated.connect_via(app)
@@ -100,23 +139,23 @@ def test_simple_signin(app, client, get_message):
 
     # Test missing choice
     data = dict(identity="matt@lp.com")
-    response = client.post("/us-signin/send-code", data=data, follow_redirects=True)
+    response = clients.post("/us-signin/send-code", data=data, follow_redirects=True)
     assert get_message("US_METHOD_NOT_AVAILABLE") in response.data
 
     # Test login using invalid email
     data = dict(identity="nobody@lp.com", chosen_method="email")
-    response = client.post("/us-signin/send-code", data=data, follow_redirects=True)
+    response = clients.post("/us-signin/send-code", data=data, follow_redirects=True)
     assert get_message("US_SPECIFY_IDENTITY") in response.data
 
     # test disabled account
     data = dict(identity="tiya@lp.com", chosen_method="email")
-    response = client.post("/us-signin/send-code", data=data, follow_redirects=True)
+    response = clients.post("/us-signin/send-code", data=data, follow_redirects=True)
     assert b"Code has been sent" not in response.data
     assert get_message("DISABLED_ACCOUNT") in response.data
 
     with capture_send_code_requests() as requests:
         with app.mail.record_messages() as outbox:
-            response = client.post(
+            response = clients.post(
                 "/us-signin/send-code",
                 data=dict(identity="matt@lp.com", chosen_method="email"),
                 follow_redirects=True,
@@ -127,7 +166,7 @@ def test_simple_signin(app, client, get_message):
     assert len(outbox) == 1
 
     # try bad code
-    response = client.post(
+    response = clients.post(
         "/us-signin",
         data=dict(identity="matt@lp.com", passcode="blahblah"),
         follow_redirects=True,
@@ -135,27 +174,27 @@ def test_simple_signin(app, client, get_message):
     assert get_message("INVALID_PASSWORD_CODE") in response.data
 
     # Correct code
-    assert "remember_token" not in [c.name for c in client.cookie_jar]
-    assert "session" not in [c.name for c in client.cookie_jar]
-    response = client.post(
+    assert "remember_token" not in [c.name for c in clients.cookie_jar]
+    assert "session" not in [c.name for c in clients.cookie_jar]
+    response = clients.post(
         "/us-signin",
         data=dict(identity="matt@lp.com", passcode=requests[0]["token"]),
         follow_redirects=False,
     )
-    assert "remember_token" not in [c.name for c in client.cookie_jar]
+    assert "remember_token" not in [c.name for c in clients.cookie_jar]
     assert "email" in auths[0][1]
 
-    response = client.get("/profile", follow_redirects=False)
+    response = clients.get("/profile", follow_redirects=False)
     assert response.status_code == 200
 
-    logout(client)
-    response = client.get("/profile", follow_redirects=False)
+    logout(clients)
+    response = clients.get("/profile", follow_redirects=False)
     assert "/login?next=%2Fprofile" in response.location
 
     # login via SMS
     sms_sender = SmsSenderFactory.createSender("test")
     set_phone(app)
-    response = client.post(
+    response = clients.post(
         "/us-signin/send-code",
         data=dict(identity="matt@lp.com", chosen_method="sms"),
         follow_redirects=True,
@@ -164,20 +203,20 @@ def test_simple_signin(app, client, get_message):
     assert b"Sign In" in response.data
 
     code = sms_sender.messages[0].split()[-1].strip(".")
-    response = client.post(
+    response = clients.post(
         "/us-signin",
         data=dict(identity="matt@lp.com", passcode=code, remember=True),
         follow_redirects=True,
     )
     assert response.status_code == 200
-    assert "remember_token" in [c.name for c in client.cookie_jar]
+    assert "remember_token" in [c.name for c in clients.cookie_jar]
     assert "sms" in auths[1][1]
 
-    response = client.get("/profile", follow_redirects=False)
+    response = clients.get("/profile", follow_redirects=False)
     assert response.status_code == 200
 
-    logout(client)
-    assert "remember_token" not in [c.name for c in client.cookie_jar]
+    logout(clients)
+    assert "remember_token" not in [c.name for c in clients.cookie_jar]
 
 
 def test_simple_signin_json(app, client_nc, get_message):
@@ -196,10 +235,7 @@ def test_simple_signin_json(app, client_nc, get_message):
         assert (
             jresponse["available_methods"] == app.config["SECURITY_US_ENABLED_METHODS"]
         )
-        assert (
-            jresponse["identity_attributes"]
-            == app.config["SECURITY_USER_IDENTITY_ATTRIBUTES"]
-        )
+        assert jresponse["identity_attributes"] == get_identity_attributes(app=app)
         assert set(jresponse["code_methods"]) == {"email", "sms"}
 
         with capture_send_code_requests() as requests:
@@ -266,6 +302,74 @@ def test_simple_signin_json(app, client_nc, get_message):
     assert len(flashes) == 0
 
 
+@pytest.mark.changeable()
+def test_signin_pwd_json(app, client, get_message):
+    # Make sure us-signin accepts a normalized and original password.
+    authenticate(client)
+    headers = {"Accept": "application/json", "Content-Type": "application/json"}
+    data = dict(
+        password="password",
+        new_password="new strong password\N{ROMAN NUMERAL ONE}",
+        new_password_confirm="new strong password\N{ROMAN NUMERAL ONE}",
+    )
+    response = client.post(
+        "/change",
+        json=data,
+        headers={"Content-Type": "application/json"},
+    )
+    assert response.status_code == 200
+    logout(client)
+
+    response = client.post(
+        "/us-signin",
+        json=dict(
+            identity="matt@lp.com", passcode="new strong password\N{ROMAN NUMERAL ONE}"
+        ),
+        headers=headers,
+        follow_redirects=False,
+    )
+    assert response.status_code == 200
+    logout(client)
+
+    response = client.post(
+        "/us-signin",
+        json=dict(
+            identity="matt@lp.com",
+            passcode="new strong password\N{LATIN CAPITAL LETTER I}",
+        ),
+        headers=headers,
+        follow_redirects=False,
+    )
+    assert response.status_code == 200
+
+
+def test_us_signin_template(app, client, get_message):
+    # Check contents of email template - this uses a test template
+    # in order to check all context vars since the default template
+    # doesn't have all of them.
+    with capture_send_code_requests() as requests:
+        with app.mail.record_messages() as outbox:
+            response = client.post(
+                "/us-signin/send-code",
+                data=dict(identity="matt@lp.com", chosen_method="email"),
+                follow_redirects=True,
+            )
+            assert len(outbox) == 1
+            matcher = re.findall(r"\w+:.*", outbox[0].body, re.IGNORECASE)
+            # should be 5 - link, email, token, config item, username
+            assert matcher[1].split(":")[1] == "matt@lp.com"
+            token = matcher[2].split(":")[1]
+            assert token == requests[0]["token"]  # deprecated
+            assert token == requests[0]["login_token"]
+            assert matcher[3].split(":")[1] == app.config["SECURITY_CONFIRM_URL"]
+            assert matcher[4].split(":")[1] == "matt@lp.com"
+
+            # check link
+            link = matcher[0].split(":", 1)[1]
+            response = client.get(link, follow_redirects=True)
+            assert get_message("PASSWORDLESS_LOGIN_SUCCESSFUL") in response.data
+
+
 def test_admin_setup_user_reset(app, client_nc, get_message):
     # Test that we can setup SMS using datastore admin method, and that
     # the datastore admin reset (reset_user_access) disables it.
@@ -382,6 +486,7 @@ def test_post_already_authenticated(client, get_message):
     )
 
 
+@pytest.mark.settings(us_email_subject="Code For You")
 def test_verify_link(app, client, get_message):
     auths = []
 
@@ -398,6 +503,9 @@ def test_verify_link(app, client, get_message):
         assert response.status_code == 200
         assert b"Sign In" in response.data
 
+    assert outbox[0].recipients == ["matt@lp.com"]
+    assert outbox[0].sender == "no-reply@localhost"
+    assert outbox[0].subject == "Code For You"
     matcher = re.match(
         r".*(http://[^\s*]*).*", outbox[0].body, re.IGNORECASE | re.DOTALL
     )
@@ -543,6 +651,7 @@ def test_setup(app, client, get_message):
 
 def test_setup_email(app, client, get_message):
     # setup with email - make sure magic link isn't sent and code is.
+    # N.B. this is using the test us_instructions template
     us_authenticate(client)
     with app.mail.record_messages() as outbox:
         response = client.post("us-setup", data=dict(chosen_method="email"))
@@ -557,13 +666,12 @@ def test_setup_email(app, client, get_message):
         verify_url = matcher.group(1)
 
     # verify no magic link
-    matcher = re.match(
-        r".*(http://[^\s*]*).*", outbox[0].body, re.IGNORECASE | re.DOTALL
-    )
-    assert not matcher
-    # grab real code
-    matcher = re.match(r".*code: ([0-9]+).*", outbox[0].body, re.IGNORECASE | re.DOTALL)
-    code = matcher.group(1)
+    matcher = re.findall(r"\w+:.*", outbox[0].body, re.IGNORECASE)
+    # should be 4 - link, email, token, config item
+    assert matcher[0].split(":")[1] == "None"
+    assert matcher[1].split(":")[1] == "matt@lp.com"
+
+    code = matcher[2].split(":")[1]
     response = client.post(verify_url, data=dict(passcode=code), follow_redirects=True)
     assert response.status_code == 200
     assert get_message("US_SETUP_SUCCESSFUL") in response.data
@@ -571,7 +679,7 @@ def test_setup_email(app, client, get_message):
 
 @pytest.mark.settings(
     us_enabled_methods=["email", "sms"],
-    user_identity_attributes=["email", "us_phone_number"],
+    user_identity_attributes=UIA_EMAIL_PHONE,
     freshness=timedelta(hours=-1),
 )
 def test_setup_json(app, client_nc, get_message):
@@ -646,7 +754,7 @@ def test_setup_json(app, client_nc, get_message):
 
 @pytest.mark.settings(
     us_enabled_methods=["email", "sms"],
-    user_identity_attributes=["email", "us_phone_number"],
+    user_identity_attributes=UIA_EMAIL_PHONE,
 )
 def test_setup_json_no_session(app, client_nc, get_message):
     # Test that with normal config freshness is required so must have session.
@@ -658,6 +766,21 @@ def test_setup_json_no_session(app, client_nc, get_message):
     }
     response = client_nc.get("/us-setup", headers=headers)
     assert response.status_code == 401
+    assert response.json["response"]["reauth_required"]
+    assert "WWW-Authenticate" not in response.headers
+
+
+@pytest.mark.settings(api_enabled_methods=["basic"])
+def test_setup_basic(app, client, get_message):
+    # If using Basic Auth - always fresh so should be able to setup (not sure the
+    # use case but...)
+    headers = {
+        "Authorization": "Basic %s"
+        % base64.b64encode(b"matt@lp.com:password").decode("utf-8")
+    }
+    response = client.get("/us-setup", headers=headers)
+    assert response.status_code == 200
+    assert b"Setup Unified Sign In options" in response.data
 
 
 def test_setup_bad_token(app, client, get_message):
@@ -666,7 +789,7 @@ def test_setup_bad_token(app, client, get_message):
 
     # bogus state
     response = client.post(
-        "/us-setup/" + "not a token", json=dict(passcode=12345), headers=headers
+        "/us-setup/not a token", json=dict(passcode=12345), headers=headers
     )
     assert response.status_code == 400
     assert response.json["response"]["error"].encode("utf-8") == get_message(
@@ -675,7 +798,7 @@ def test_setup_bad_token(app, client, get_message):
 
     # same w/o json
     response = client.post(
-        "/us-setup/" + "not a token", data=dict(passcode=12345), follow_redirects=True
+        "/us-setup/not a token", data=dict(passcode=12345), follow_redirects=True
     )
     assert get_message("API_ERROR") in response.data
 
@@ -780,7 +903,9 @@ def test_verify_json(app, client, get_message):
     }
 
     response = client.post(
-        "us-verify/send-code", json=dict(chosen_method="orb"), headers=headers,
+        "us-verify/send-code",
+        json=dict(chosen_method="orb"),
+        headers=headers,
     )
     assert response.status_code == 400
 
@@ -788,7 +913,9 @@ def test_verify_json(app, client, get_message):
     sms_sender = SmsSenderFactory.createSender("test")
     set_phone(app)
     response = client.post(
-        "us-verify/send-code", json=dict(chosen_method="sms"), headers=headers,
+        "us-verify/send-code",
+        json=dict(chosen_method="sms"),
+        headers=headers,
     )
     assert response.status_code == 200
 
@@ -885,7 +1012,7 @@ def test_invalid_method_setup(app, client, get_message):
 
 def test_setup_new_totp(app, client, get_message):
     # us-setup should generate a new totp secret for each setup
-    # Verify  existing cods no longer work
+    # Verify existing codes no longer work
     us_authenticate(client)
 
     headers = {"Accept": "application/json", "Content-Type": "application/json"}
@@ -898,6 +1025,7 @@ def test_setup_new_totp(app, client, get_message):
         headers=headers,
     )
     assert response.status_code == 200
+    assert "authr_key" not in response.json["response"]
     code = sms_sender.messages[0].split()[-1].strip(".")
 
     # Now start setup again - it should generate a new totp - so the previous 'code'
@@ -927,31 +1055,34 @@ def test_setup_new_totp(app, client, get_message):
 
 
 def test_qrcode(app, client, get_message):
-    headers = {"Accept": "application/json", "Content-Type": "application/json"}
+    # Test forms based authenticator setup - can't really parse QRcode - but can use
+    # the key sent as part of the response.
     us_authenticate(client, identity="gal@lp.com")
-    response = client.post(
-        "us-setup", json=dict(chosen_method="authenticator"), headers=headers
-    )
+    response = client.post("us-setup", data=dict(chosen_method="authenticator"))
     assert response.status_code == 200
-    state = response.json["response"]["state"]
+    # verify png QRcode is present
+    assert b"data:image/svg+xml;base64," in response.data
 
-    # Now request code. We can't test the qrcode easily - but we can get the totp_secret
-    # that goes into the qrcode and make sure that works
-    mtf = Mock(wraps=app.security._totp_factory)
-    app.security.totp_factory(mtf)
-    qrcode_page_response = client.get("/us-qrcode/" + state, follow_redirects=True)
-    assert mtf.get_totp_uri.call_count == 1
-    (username, totp_secret), _ = mtf.get_totp_uri.call_args
-    assert username == "gal@lp.com"
-    assert b"svg" in qrcode_page_response.data
+    # parse out key
+    rd = response.data.decode("utf-8")
+    matcher = re.match(r".*((?:\S{4}-){7}\S{4}).*", rd, re.DOTALL)
+    totp_secret = matcher.group(1)
 
     # Generate token from passed totp_secret and confirm setup
-    code = app.security._totp_factory.generate_totp_password(totp_secret)
-    response = client.post(
-        "/us-setup/" + state, json=dict(passcode=code), headers=headers
+    totp = TOTP(totp_secret)
+    code = totp.generate().token
+
+    # get verify link e.g. /us-setup/{state}
+    matcher = re.match(
+        r'.*<form action="([^\s]*)".*',
+        response.data.decode("utf-8"),
+        re.IGNORECASE | re.DOTALL,
     )
+    verify_url = matcher.group(1)
+
+    response = client.post(verify_url, data=dict(passcode=code), follow_redirects=True)
     assert response.status_code == 200
-    assert response.json["response"]["chosen_method"] == "authenticator"
+    assert get_message("US_SETUP_SUCCESSFUL") in response.data
 
 
 def test_next(app, client, get_message):
@@ -981,6 +1112,30 @@ def test_next(app, client, get_message):
     assert response.location == "http://localhost/post_login"
 
 
+@pytest.mark.registerable()
+@pytest.mark.confirmable()
+@pytest.mark.settings(requires_confirmation_error_view="/confirm")
+def test_requires_confirmation_error_redirect(app, client):
+    data = dict(
+        email="jyl@lp.com", password="password", password_confirm="password", next=""
+    )
+    response = client.post("/register", data=data, follow_redirects=True)
+
+    with capture_send_code_requests() as requests:
+        response = client.post(
+            "/us-signin/send-code",
+            data=dict(identity="jyl@lp.com", chosen_method="email"),
+            follow_redirects=True,
+        )
+
+    response = client.post(
+        "/us-signin",
+        data=dict(identity="jyl@lp.com", passcode=requests[0]["token"]),
+        follow_redirects=True,
+    )
+    assert b"send_confirmation_form" in response.data
+
+
 @pytest.mark.registerable()
 @pytest.mark.confirmable()
 def test_confirmable(app, client, get_message):
@@ -1051,8 +1206,10 @@ def test_can_add_password(app, client, get_message):
 
 @pytest.mark.settings(
     us_enabled_methods=["password"],
-    user_identity_attributes=["email", "username"],
-    user_identity_mappings=[{"email": uia_email_mapper}, {"username": lambda x: x}],
+    user_identity_attributes=[
+        {"email": {"mapper": uia_email_mapper}},
+        {"username": {"mapper": lambda x: x}},
+    ],
 )
 def test_regular_login(app, client, get_message):
     # If "password" in methods - then should be able to login with good-ol
@@ -1079,7 +1236,7 @@ def test_regular_login(app, client, get_message):
 
 
 @pytest.mark.settings(
-    us_enabled_methods=["sms"], user_identity_attributes=["email", "us_phone_number"]
+    us_enabled_methods=["sms"], user_identity_attributes=UIA_EMAIL_PHONE
 )
 def test_regular_login_disallowed(app, client, get_message):
     # If "password" not in methods - then should not be able to use password
@@ -1172,7 +1329,7 @@ def test_tf_link_spa(app, client, get_message):
 
 @pytest.mark.two_factor()
 @pytest.mark.settings(
-    two_factor_required=True, user_identity_attributes=["email", "us_phone_number"]
+    two_factor_required=True, user_identity_attributes=UIA_EMAIL_PHONE
 )
 def test_tf_not(app, client, get_message):
     # Test basic two-factor - when first factor doesn't require second (e.g. SMS)
@@ -1320,7 +1477,9 @@ def test_passwd_and_authenticator(app, client, get_message):
     response = client.post("us-setup", data=dict(chosen_method="authenticator"))
     assert response.status_code == 200
     assert b"Code has been sent" not in response.data
-    assert b"Open your authenticator app" in response.data
+    assert b"Open an authenticator app" in response.data
+    # verify png QRcode is present
+    assert b"data:image/svg+xml;base64," in response.data
 
     headers = {"Accept": "application/json", "Content-Type": "application/json"}
     response = client.get("us-setup", headers=headers)
@@ -1347,18 +1506,17 @@ def test_totp_generation(app, client, get_message):
         "us-setup", json=dict(chosen_method="authenticator"), headers=headers
     )
     assert response.status_code == 200
+    assert response.json["response"]["authr_issuer"] == "service_name"
+    assert response.json["response"]["authr_username"] == "dave@lp.com"
+    assert "authr_key" in response.json["response"]
+
     state = response.json["response"]["state"]
 
-    # Now request code. We can't test the qrcode easily - but we can get the totp_secret
-    # that goes into the qrcode and make sure that works
-    mtf = Mock(wraps=app.security._totp_factory)
-    app.security.totp_factory(mtf)
-    client.get("/us-qrcode/" + state, follow_redirects=True)
-    assert mtf.get_totp_uri.call_count == 1
-    (username, totp_secret), _ = mtf.get_totp_uri.call_args
-
     # Generate token from passed totp_secret and confirm setup
-    code = app.security._totp_factory.generate_totp_password(totp_secret)
+    totp_key = response.json["response"]["authr_key"]
+    totp = TOTP(totp_key)
+    code = totp.generate().token
+
     response = client.post(
         "/us-setup/" + state, json=dict(passcode=code), headers=headers
     )
@@ -1369,37 +1527,10 @@ def test_totp_generation(app, client, get_message):
     with app.app_context():
         user = app.security.datastore.find_user(email="dave@lp.com")
         ts = app.security.datastore.us_get_totp_secrets(user)
-        assert totp_secret == ts["authenticator"]
-
-    # Ok - setup again, this time don't verify and
-    # make sure totp_secret in DB doesn't change
-    response = client.post(
-        "us-setup", json=dict(chosen_method="authenticator"), headers=headers
-    )
-    assert response.status_code == 200
-    state = response.json["response"]["state"]
-
-    # Now request code. We can't test the qrcode easily - but we can get the totp_secret
-    # that goes into the qrcode and make sure that works
-    mtf = Mock(wraps=app.security._totp_factory)
-    app.security.totp_factory(mtf)
-    client.get("/us-qrcode/" + state, follow_redirects=True)
-    assert mtf.get_totp_uri.call_count == 1
-    (username, new_totp_secret), _ = mtf.get_totp_uri.call_args
-    # Should generate a new totp secret for each setup
-    assert totp_secret != new_totp_secret
-
-    # validate with wrong code
-    response = client.post(
-        "/us-setup/" + state, json=dict(passcode=123345), headers=headers
-    )
-    assert response.status_code == 400
-
-    # Make sure totp_secret in DB didn't change
-    with app.app_context():
-        user = app.security.datastore.find_user(email="dave@lp.com")
-        ts = app.security.datastore.us_get_totp_secrets(user)
-        assert totp_secret == ts["authenticator"]
+        assert (
+            app.security._totp_factory.get_totp_pretty_key(ts["authenticator"])
+            == totp_key
+        )
 
     # Now setup SMS and verify that authenticator totp hasn't changed
     sms_sender = SmsSenderFactory.createSender("test")
@@ -1422,4 +1553,89 @@ def test_totp_generation(app, client, get_message):
     with app.app_context():
         user = app.security.datastore.find_user(email="dave@lp.com")
         ts = app.security.datastore.us_get_totp_secrets(user)
-        assert totp_secret == ts["authenticator"]
+        assert (
+            app.security._totp_factory.get_totp_pretty_key(ts["authenticator"])
+            == totp_key
+        )
+
+    # Ok - setup again - but send invalid code - check totp in DB didn't change.
+    response = client.post(
+        "us-setup", json=dict(chosen_method="authenticator"), headers=headers
+    )
+    assert response.status_code == 200
+    state = response.json["response"]["state"]
+    totp_key = response.json["response"]["authr_key"]
+
+    # validate with wrong code
+    response = client.post(
+        "/us-setup/" + state, json=dict(passcode=123345), headers=headers
+    )
+    assert response.status_code == 400
+
+    # Make sure totp_secret in DB didn't change
+    with app.app_context():
+        user = app.security.datastore.find_user(email="dave@lp.com")
+        ts = app.security.datastore.us_get_totp_secrets(user)
+        assert (
+            app.security._totp_factory.get_totp_pretty_key(ts["authenticator"])
+            != totp_key
+        )
+
+
+@pytest.mark.two_factor()
+@pytest.mark.settings(
+    two_factor_required=True,
+    user_identity_attributes=UIA_EMAIL_PHONE,
+    two_factor_always_validate=False,
+)
+def test_us_tf_validity(app, client, get_message):
+    us_tf_authenticate(app, client, remember=True)
+    logout(client)
+    data = dict(identity="gal@lp.com", passcode="password")
+    response = client.post(
+        "/us-signin", json=data, headers={"Content-Type": "application/json"}
+    )
+    assert b'"code": 200' in response.data
+    cookie = next(
+        (cookie for cookie in client.cookie_jar if cookie.name == "tf_validity"), None
+    )
+    assert cookie is not None
+
+    logout(client)
+
+    data = dict(identity="gal2@lp.com", passcode="password")
+    response = client.post("/us-signin", data=data, follow_redirects=True)
+    assert b"Please enter your authentication code" in response.data
+
+    # clear the cookie to make sure it's not picking it up with json.
+    client.cookie_jar.clear("localhost.local", "/", "tf_validity")
+
+    token = us_tf_authenticate(app, client, remember=True, json=True)
+    logout(client)
+    data = dict(identity="gal@lp.com", passcode="password", tf_validity_token=token)
+    response = client.post(
+        "/us-signin",
+        json=data,
+        follow_redirects=True,
+        headers={"Content-Type": "application/json"},
+    )
+    assert response.status_code == 200
+    # verify logged in
+    response = client.get("/profile", follow_redirects=False)
+
+    assert response.status_code == 200
+    assert b"Welcome gal@lp.com" in response.data
+
+    logout(client)
+
+    data["identity"] = "gal2@lp.com"
+    response = client.post(
+        "/us-signin",
+        json=data,
+        follow_redirects=True,
+        headers={"Content-Type": "application/json"},
+    )
+    assert response.status_code == 200
+    assert response.json["response"]["tf_primary_method"] == "authenticator"
+    assert response.json["response"]["tf_required"]
+    assert response.json["response"]["tf_state"] == "ready"
